import logging
import os
import sys

import pandas as pd
import geopandas
import numpy as np
import math

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

import matplotlib

from matplotlib import cm
from matplotlib.colors import ListedColormap
import matplotlib.font_manager as font_manager
import matplotlib.patheffects as path_effects

import contextily as cx

import matplotlib.patches as mpatches

logger = logging.getLogger('PT3S')  

def cmpTIMEs( df=pd.DataFrame() # a V3sErg=dx.MxAdd(mx) df i.e. v3sKNOT=V3sErg['V3_KNOT']
             ,col='KNOT~*~*~*~PH'
             ,timeLstA=[]
             ,timeLstB=[]             
             ,newColNamesBase=[]):

    """
    compares the value of col between 2 TIMEs (B-A) and creates new cols    
    """

    pass
    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

    try:
        for timeA,timeB,newColNameBase in zip(timeLstA,timeLstB,newColNamesBase):

             ergCol_timeA=('TIME'
            ,col
            ,pd.Timestamp(timeA.strftime('%Y-%m-%d %X.%f'))
            ,pd.Timestamp(timeA.strftime('%Y-%m-%d %X.%f'))
            )

             ergCol_timeB=('TIME'
            ,col
            ,pd.Timestamp(timeB.strftime('%Y-%m-%d %X.%f'))
            ,pd.Timestamp(timeB.strftime('%Y-%m-%d %X.%f'))
            )

             if ergCol_timeA not in df.columns:               
                logger.warning("{:s}col: {:s} nicht vorhanden.".format(logStr,col)) 
                continue
             if ergCol_timeB not in df.columns:
                logger.warning("{:s}col: {:s} nicht vorhanden.".format(logStr,col)) 
                continue
             
             df[newColNameBase+'_DIF']=df.apply(lambda row: row[ergCol_timeB]-row[ergCol_timeA] ,axis=1)    
             df[newColNameBase+'_DIFAbs']=df.apply(lambda row: math.fabs(row[newColNameBase+'_DIF']) ,axis=1)    
         
         
    except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise e         
    finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

def pNFD_FW(                   
                 ax=None
                ,axTitle=None
                
                ,gdf_ROHR=geopandas.GeoDataFrame()      
                ,gdf_FWVB=geopandas.GeoDataFrame()         
                ,gdf_KNOT=geopandas.GeoDataFrame()                                 
             
                # Layout ROHR
            
                # Sachdatum
                ,attr_colors_ROHR_Sach = 'DI' 
                ,attr_colors_ROHR_Sach_zOrder = 3
                ,colors_ROHR_Sach = ['lightgray', 'dimgray']                
                ,norm_min_ROHR_Sach = None 
                ,norm_max_ROHR_Sach = None

                ,attr_lws_ROHR_Sach='DI'                     
                ,attr_colors_ROHR_Sach_patches_fmt="DN (Innen) {:4.0f}"
                ,attr_colors_ROHR_Sach_patchValues=None
                
                # Breitenfaktor (fuer Sach- und Ergebnisdaten)
                ,fac_lws_ROHR=5. 

                # Ergebnis: Farbe
                ,attr_colors_ROHR_Erg=None 
                ,attr_colors_ROHR_Erg_zOrder = 4 
                ,colors_ROHR_Erg = ['darkgreen','magenta']
                ,norm_min_ROHR_Erg = None 
                ,norm_max_ROHR_Erg = None            
                ,attr_colors_ROHR_Erg_patches_fmt="Q (abs.) {:4.0f} t/h"
                ,attr_colors_ROHR_Erg_patchValues=None
                
                                
                ,query_ROHR_Erg=None #'{:s} > 1.'.format(attr_colors_ROHR_Erg)
                # wenn angegeben, werden nur Ergebnisse von Rohren geplottet die dem query entsprechen
                                
                # Ergebnis: Breite: Standard: so wie die Sachdaten (diese werden dann komplett ueberzeichnet) ...
                ,lws_ROHR_Erg_Sach=True
                # andernfalls richtet sich die Breite nach dem Ergebnisattribut; es ergeben sich _dann in absteigender Richtung duennere oder dickere Breiten als bei den Sachdaten 
                # _Bsp. Sach DI und Erg QAbs: der max. DN und der max. Fluss werden mitderselben Breite gezeichnet; darunter ergeben sich unterschiedliche Breiten; DN i.d.R. breiter als Fluss
                ,fac_lws_ROHR_Erg=None # nur wirksam bei lws_ROHR_Erg_Sach=False; dann Standard fac_lws_ROHR  
            

                # Layout FWVB
            
                # Sachdatum
                ,attr_colors_FWVB_Sach='W0'
                ,attr_colors_FWVB_Sach_zOrder = 1
                ,attr_colors_FWVB_Sach_patches_fmt="W {:4.0f} kW"        
                ,attr_colors_FWVB_Sach_patchValues=None
                ,colors_FWVB_Sach = ['oldlace', 'orange'] 
                ,norm_min_FWVB_Sach = None 
                ,norm_max_FWVB_Sach = None            
            
                ,attr_ms_FWVB_Sach='W0' 
                ,fac_ms_FWVB=None # fac_ms_KNOT oder 8000.  wenn beides undefiniert

                # Ergebnis: Farbe
                ,attr_colors_FWVB_Erg=None 
                ,attr_colors_FWVB_Erg_zOrder = 2
                ,attr_colors_FWVB_Erg_patches_fmt="dp {:4.1f} bar"     
                ,attr_colors_FWVB_Erg_patchValues = None
                ,attr_colors_FWVB_ErgNeg_patches_fmt="dp -{:4.1f} bar"                     
                ,colors_FWVB_Erg = ['aquamarine','teal'] 
                ,norm_min_FWVB_Erg = None 
                ,norm_max_FWVB_Erg = None         
                
                # Ergebnis: Groesse: Standard: -1
                ,ms_FWVB_Erg=-1
                # -2: konst. Groesse 
                # -1: Groesse so wie die Sachdaten (diese werden dann komplett ueberzeichnet) 
                #       sind die Sachdaten nicht angefordert dasselbe Ergebnis wie ms_FWVB_Erg=0
                #  0: Groesse nach Ergebnisdaten (bei dp als Erg und W0 als Sach fuehrt dies zu vielen Überzeichnungen)
                # >0: Prozent (1=100%): Groesse nach Ergebnisdaten, aber nur max. ms_FWVB_Erg-Prozent der Sachdatengroesse 
                #               dies führt bei ]0;1[ dazu, dass die Sachdatenannotation immer sichtbar bleibt

                ,colors_FWVB_ErgNeg = ['lightskyblue','royalblue'] # ['aquamarine','skyblue']
                # wenn nicht None, dann werden negative Werte mit dieser Farbe gezeichnet; beide Farbskalen (die dann pos. und diese neg.) werden voll ausgenutzt
                # die Groesse erstreckt sich ueber den gesamten Wertebereich d.h. z.B. -.25 ist so gross wie +.25                

            
                # Layout NODES (KNOTen)
                ,attr_colors_KNOT_Erg=None 
                ,attr_colors_KNOT_Erg_patches_fmt="pDiff +{:4.2f} bar"        
                ,attr_colors_KNOT_ErgNeg_patches_fmt="pDiff -{:4.2f} bar"  
                ,attr_colors_KNOT_Erg_patchValues=None                         
                ,colors_KNOT_Erg = ['yellow','red']
                ,marker_KNOT_Erg='.'
                ,norm_min_KNOT_Erg = None 
                ,norm_max_KNOT_Erg = None  
                
                ,size_min_KNOT_Erg=None      
                ,size_max_KNOT_Erg=None  
                
                ,attr_colors_KNOT_Erg_zOrder=5

                ,colors_KNOT_ErgNeg = ['yellowgreen','sienna']
                # wenn nicht None, dann werden negative Werte mit dieser Farbe gezeichnet
                # norm_max_KNOT_Erg nicht vorgegeben:
                #   beide Farbskalen (die pos. und diese neg.) werden voll ausgenutzt
                #   die Groesse erstreckt sich ueber den Absolutwert d.h. z.B. -.25 ist so gross wie +.25 

                ,fac_ms_KNOT = None  # fac_ms_FWVB oder 8000.  wenn beides undefiniert

                # Layout Karte
                ,MapOn=False
                ,LabelsOnTop = False                               
                ,Map_resolution = 15

                ):
    
        """
        Plots geopandas-Dfs: gdf_ROHR, gdf_FWVB, gdf_KNOT.
       
        
        Args:
            gdf_ROHR (gdf, optional, default=empty gdf)         
            gdf_FWVB (gdf, optional, default=empty gdf)    
            gdf_KNOT (gdf, optional, default=empty gdf)      
            
            Layout KNOT (Nodes):
                attr_colors_KNOT_Erg (col in gdf_KNOT, default=None)                          
                colors_KNOT_Erg (LinearSegmentedColormap.from_list(...), default=['yellow','red'])
                fac_ms_KNOT (factor for markersize, default=None ==> fac_ms_FWVB or 8.000 if fac_ms_FWVB==None)
                
                gdf_KNOT[attr_colors_KNOT_Erg].min(): markersize=0 * fac_ms_KNOT/'yellow'
                gdf_KNOT[attr_colors_KNOT_Erg].max(): markersize=1 * fac_ms_KNOT/'red'
                
                norm_min_KNOT_Erg (default=None):
                    gdf_KNOT[attr_colors_KNOT_Erg]<=norm_min_KNOT_Erg markersize<=0 * fac_ms_KNOT/'yellow'                    
                norm_max_KNOT_Erg (default=None)        
                    gdf_KNOT[attr_colors_KNOT_Erg]>=norm_max_KNOT_Erg markersize>=1 * fac_ms_KNOT/'red'  
                    
                size_min_KNOT_Erg (]0,1[, default=None):
                    markersize>= size_min_KNOT_Erg * fac_ms_KNOT 
                size_max_KNOT_Erg (]0,1[, default=None):        
                    markersize<= size_max_KNOT_Erg * fac_ms_KNOT
                    
                attr_colors_KNOT_Erg_zOrder (default=5; ROHR_Sach=3,ROHR_Erg=4,FWVB_Sach=1,FWVB_Erg=2) 
                
                marker_KNOT_Erg (default='.')
                
                attr_colors_KNOT_Erg_patchValues (default=None)
                attr_colors_KNOT_Erg_patches_fmt (default='pDiff +{:4.2f} bar')        
                
                
             #        
            
             ,colors_KNOT_ErgNeg = ['yellowgreen','sienna']
             # wenn nicht None, dann werden negative Werte mit dieser Farbe gezeichnet
             # norm_max_KNOT_Erg nicht vorgegeben:
             #   beide Farbskalen (die pos. und diese neg.) werden voll ausgenutzt
             #   die Groesse erstreckt sich ueber den Absolutwert d.h. z.B. -.25 ist so gross wie +.25 
            
               
             
            
             ,attr_colors_KNOT_ErgNeg_patches_fmt="pDiff -{:4.2f} bar"               
                
        Returns:            
             attr_colors_ROHR_Sach_patches: Legendeneinträge
             attr_colors_ROHR_Erg_patches: Legendeneinträge     
             attr_colors_FWVB_Sach_patches: Legendeneinträge
             attr_colors_FWVB_Erg_patches: Legendeneinträge    
             attr_colors_KNOT_Erg_patches: Legendeneinträge    
        
        """    
    


        logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            
            attr_colors_ROHR_Sach_patches=None
            attr_colors_ROHR_Erg_patches=None
            attr_colors_FWVB_Sach_patches=None
            attr_colors_FWVB_Erg_patches=None
            attr_colors_KNOT_Erg_patches=None
           
            if ax == None:
                ax = plt.gca()

            ### ROHRe ###
        
            attr_colors_ROHR_Sach_patches=[]
            attr_colors_ROHR_Erg_patches=[]       
            if not gdf_ROHR.empty:        
        
        
                # Erstellen der Colormap 
                cmap_ROHR = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_ROHR', colors_ROHR_Sach, N = 256)        
            
                # Normierung fuer Colormap
                if norm_min_ROHR_Sach == None:            
                    norm_min_ROHR_Sach=gdf_ROHR[attr_colors_ROHR_Sach].min()
                if norm_max_ROHR_Sach == None:
                    norm_max_ROHR_Sach=gdf_ROHR[attr_colors_ROHR_Sach].max()
                    
                norm_diff_ROHR_Sach=norm_max_ROHR_Sach-norm_min_ROHR_Sach
                
                if norm_diff_ROHR_Sach < 0.01:
                    norm_min_ROHR_Sach=0.99*norm_max_ROHR_Sach
                    norm_diff_ROHR_Sach=norm_max_ROHR_Sach-norm_min_ROHR_Sach
                                
                logger.debug("{0:s}norm_min_ROHR_Sach: {1:10.2f} norm_max_ROHR_Sach: {2:10.2f}".format(logStr,norm_min_ROHR_Sach,norm_max_ROHR_Sach)) 
                                
                norm_ROHR_color = plt.Normalize(vmin=norm_min_ROHR_Sach, vmax=norm_max_ROHR_Sach) 
                                
                # Plotten ROHRe
                gdf_ROHR.plot(ax = ax
                             ,zorder = attr_colors_ROHR_Sach_zOrder
                             ,linewidth = norm_ROHR_color(gdf_ROHR[attr_lws_ROHR_Sach]) * fac_lws_ROHR
                             # wg. astype: ufunc 'isnan' not supported for the input types, and the inputs could not be safely coerced to any supported types according to the casting rule ''safe''
                             ,color = cmap_ROHR(norm_ROHR_color(gdf_ROHR[attr_colors_ROHR_Sach].astype(float)))
                             ,path_effects=[path_effects.Stroke(capstyle="round")]
                             )
            
                if attr_colors_ROHR_Sach_patchValues == None:
                     attr_colors_ROHR_Sach_patchValues=np.arange(norm_min_ROHR_Sach,norm_max_ROHR_Sach+1,norm_diff_ROHR_Sach/4)                 
                attr_colors_ROHR_Sach_patches = [ mpatches.Patch(color=cmap_ROHR(norm_ROHR_color(value)), label=attr_colors_ROHR_Sach_patches_fmt.format(value) ) 
                                                 for value in attr_colors_ROHR_Sach_patchValues]        
            
                    
                if attr_colors_ROHR_Erg != None:
                    
                    # Erstellen der Colormap 
                    cmap_ROHRErg = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_ROHRErg', colors_ROHR_Erg, N = 256)     
                
                    # Normierung fuer Colormap
                    if norm_min_ROHR_Erg == None:            
                        norm_min_ROHR_Erg=gdf_ROHR[attr_colors_ROHR_Erg].min()    
                    if norm_max_ROHR_Erg == None:            
                        norm_max_ROHR_Erg=gdf_ROHR[attr_colors_ROHR_Erg].max()
                    norm_ROHRErg_color = plt.Normalize(vmin=norm_min_ROHR_Erg, vmax=norm_max_ROHR_Erg)       
                    norm_diff_ROHR_Erg=norm_max_ROHR_Erg-norm_min_ROHR_Erg
                
                    # Breite
                    if lws_ROHR_Erg_Sach:            
                        attr_lws_ROHR_Erg=attr_colors_ROHR_Sach
                        norm_ROHR_lws_Erg=norm_ROHR_color
                        fac_lws_ROHR_Erg=fac_lws_ROHR
                    else:
                        attr_lws_ROHR_Erg=attr_colors_ROHR_Erg
                        norm_ROHR_lws_Erg=norm_ROHRErg_color
                        
                        if fac_lws_ROHR_Erg == None:
                            fac_lws_ROHR_Erg=fac_lws_ROHR
                    
                    if query_ROHR_Erg != None:
                        df=gdf_ROHR.query(query_ROHR_Erg)
                    else:
                        df=gdf_ROHR
                    # große ueber kleine zeichnen
                    df=df.sort_values(by=[attr_colors_ROHR_Erg],ascending=True)
            
                    df.plot(ax = ax
                             ,zorder = attr_colors_ROHR_Erg_zOrder
                             ,linewidth = norm_ROHR_lws_Erg(df[attr_lws_ROHR_Erg]) * fac_lws_ROHR_Erg
                             ,color = cmap_ROHRErg(norm_ROHRErg_color(df[attr_colors_ROHR_Erg]))
                             ,path_effects=[path_effects.Stroke(capstyle="round")]
                             )       
                    
                    
                    # Legendeneinräge
                                 
                    if attr_colors_ROHR_Erg_patchValues == None:
                         attr_colors_ROHR_Erg_patchValues=np.arange(norm_min_ROHR_Erg,norm_max_ROHR_Erg+1,norm_diff_ROHR_Erg/4)                 
                    attr_colors_ROHR_Erg_patches = [ mpatches.Patch(color=cmap_ROHRErg(norm_ROHRErg_color(value)), label=attr_colors_ROHR_Erg_patches_fmt.format(value) ) 
                                                     for value in attr_colors_ROHR_Erg_patchValues]                                                                                                                                 
                                                                                                                              
            
            ### FWVBe ###
            attr_colors_FWVB_Sach_patches=[]
            attr_colors_FWVB_Erg_patches=[]       
            if not gdf_FWVB.empty:

                # Groessenfaktor
                if fac_ms_FWVB == None:
                    if fac_ms_KNOT != None:
                        fac_ms_FWVB=fac_ms_KNOT
                    else:
                        fac_ms_FWVB=8000.
                        
                        
                logger.debug("{0:s}fac_ms_FWVB: {1:10.2f}".format(logStr,fac_ms_FWVB)) 

                #attr_colors_FWVB_Sach_patches=[]
                if attr_colors_FWVB_Sach != None:

                    # Erstellen der Colormap
                    cmap_FWVB = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_FWVB', colors_FWVB_Sach, N = 256)        

                    # Normierung fuer Colormap
                    if norm_min_FWVB_Sach == None:            
                        norm_min_FWVB_Sach=gdf_FWVB[attr_colors_FWVB_Sach].min()
                    if norm_max_FWVB_Sach == None:
                        norm_max_FWVB_Sach=gdf_FWVB[attr_colors_FWVB_Sach].max()
                        
                    #norm_FWVB_color = plt.Normalize(vmin=norm_min_FWVB_Sach, vmax=norm_max_FWVB_Sach)
                    # Werte außerhalb vmin, vmax erhalten ohne clip Werte < 0 > 1 und damit (bei Wert <= 0) keine Darstellung wenn die Normierung auch fuer die Groesse verwendet wird
                    norm_diff_FWVB_Sach=norm_max_FWVB_Sach-norm_min_FWVB_Sach
                    
                    if norm_diff_FWVB_Sach < 0.01:
                        norm_min_FWVB_Sach=0.99*norm_max_FWVB_Sach
                        norm_diff_FWVB_Sach=norm_max_FWVB_Sach-norm_min_FWVB_Sach      
                        
                        
                    logger.debug("{0:s}norm_min_FWVB_Sach: {1:10.2f} norm_max_FWVB_Sach: {2:10.2f}".format(logStr,norm_min_FWVB_Sach,norm_max_FWVB_Sach)) 
                    norm_FWVB_color = plt.Normalize(vmin=norm_min_FWVB_Sach, vmax=norm_max_FWVB_Sach)                    
                    
                    logger.debug("{0:s}ms_min_FWVB_Sach: {1:10.2f} ms_max_FWVB_Sach: {2:10.2f}".format(logStr
                                                                                                         ,norm_FWVB_color(gdf_FWVB[attr_ms_FWVB_Sach].min())
                                                                                                         ,norm_FWVB_color(gdf_FWVB[attr_ms_FWVB_Sach].max())
                                 )) 
                
                    # Plotten FWVB
                    gdf_FWVB.plot(ax = ax
                                ,zorder = attr_colors_FWVB_Sach_zOrder 
                                ,marker = '.'
                                ,markersize = norm_FWVB_color(gdf_FWVB[attr_ms_FWVB_Sach]) * fac_ms_FWVB                       
                                ,color = cmap_FWVB(norm_FWVB_color(gdf_FWVB[attr_colors_FWVB_Sach]))
                                )
                            
                    if attr_colors_FWVB_Sach_patchValues == None:
                         attr_colors_FWVB_Sach_patchValues=np.arange(norm_min_FWVB_Sach,norm_max_FWVB_Sach+1,norm_diff_FWVB_Sach/4)    
                         
                    attr_colors_FWVB_Sach_patches = [ mpatches.Patch(color=cmap_FWVB(norm_FWVB_color(value)), label=attr_colors_FWVB_Sach_patches_fmt.format(value) ) 
                                                     for value in attr_colors_FWVB_Sach_patchValues]                                                                                                                                 
                                                                                                                               
                #attr_colors_FWVB_Erg_patches=[]
                if attr_colors_FWVB_Erg != None:
                    
                    logger.debug("{0:s}attr_colors_FWVB_Erg: {1:s}".format(logStr,str(attr_colors_FWVB_Erg))) 

                    minValue=gdf_FWVB[attr_colors_FWVB_Erg].astype(float).min()
                
                    # Erstellen der Colormaps 
                    cmap_FWVBErg = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_FWVBErg', colors_FWVB_Erg, N = 256)     

                    if colors_FWVB_ErgNeg != None and minValue <0:                    
                        cmap_FWVBErgNeg = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_FWVBErgNeg', colors_FWVB_ErgNeg, N = 256)  
            
                    # Normierung fuer (1) Colormap und Groesse
                    if norm_min_FWVB_Erg == None:    
                         if colors_FWVB_ErgNeg != None and minValue <0:
                            norm_min_FWVB_Erg=0.
                         else:
                            norm_min_FWVB_Erg=gdf_FWVB[attr_colors_FWVB_Erg].astype(float).min()    
                    if norm_max_FWVB_Erg == None:            
                        if colors_FWVB_ErgNeg != None and minValue <0:
                            norm_max_FWVB_Erg=max(gdf_FWVB[attr_colors_FWVB_Erg].astype(float).max(),-gdf_FWVB[attr_colors_FWVB_Erg].astype(float).min())
                        else:
                            norm_max_FWVB_Erg=gdf_FWVB[attr_colors_FWVB_Erg].astype(float).max()
                    norm_FWVBErg_color = plt.Normalize(vmin=norm_min_FWVB_Erg, vmax=norm_max_FWVB_Erg)    
                    norm_diff_FWVB_Erg=norm_max_FWVB_Erg-norm_min_FWVB_Erg
                
                    # Groesse
                    if ms_FWVB_Erg == -2:                           
                        markersizes=1. * fac_ms_FWVB                       
                    elif ms_FWVB_Erg == -1:   
                        if attr_colors_FWVB_Sach != None:
                            markersizes=norm_FWVB_color(gdf_FWVB[attr_colors_FWVB_Sach]) * fac_ms_FWVB       
                        else:
                            # Sachdaten nicht definiert
                            try:
                                markersizes=norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float)) * fac_ms_FWVB           
                            except ValueError:
                                logger.debug("{0:s}ms_FWVB_Erg: {1:d} und attr_colors_FWVB_Sach nicht definiert; Verhalten wie bei -2 ...".format(logStr,ms_FWVB_Erg)) 
                                markersizes=1. * fac_ms_FWVB   
                                pass
                                
                            
                    elif ms_FWVB_Erg == 0:                         
                        markersizes=norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float)) * fac_ms_FWVB           

                    else:
                        if attr_colors_FWVB_Sach != None:
                            markersizesSach=norm_FWVB_color(gdf_FWVB[attr_colors_FWVB_Sach]) * fac_ms_FWVB                
                        else:
                            markersizesSach=norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float)) * fac_ms_FWVB          
                        markersizesErg=norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float)) * fac_ms_FWVB           
                        markersizes=[min(ms_FWVB_Erg*msSach,msErg) for (msSach,msErg) in zip(markersizesSach,markersizesErg)]
                      
                    #if isinstance(markersizes,list):    
                    #    logger.debug("{0:s}markersizes: min {1:10.2f} max {2:10.2f} ".format(logStr,min(markersizes),max(markersizes))) 
                    #else:
                    #    logger.debug("{0:s}markersizes: {1:10.2f}  ".format(logStr,markersizes)) 

                    # Plotten FWVB

                    if colors_FWVB_ErgNeg != None and minValue <0:
                        # Farbnormierungen
                        norm_FWVBErg_color=plt.Normalize(vmin=norm_min_FWVB_Erg, vmax=gdf_FWVB[attr_colors_FWVB_Erg].astype(float).max())  
                        norm_FWVBErgNeg_color=plt.Normalize(vmin=norm_min_FWVB_Erg, vmax=-gdf_FWVB[attr_colors_FWVB_Erg].astype(float).min())  

                        norm_FWVBErg_size=plt.Normalize(vmin=0., vmax=max(gdf_FWVB[attr_colors_FWVB_Erg].astype(float).max(),-gdf_FWVB[attr_colors_FWVB_Erg].astype(float).min()))  

                        gdf=gdf_FWVB[gdf_FWVB[attr_colors_FWVB_Erg].astype(float)>=0]
                        if not gdf.empty:
                            gdf.plot(ax = ax
                                        ,zorder = attr_colors_FWVB_Erg_zOrder  
                                        ,marker = '.'                           
                                        ,markersize = norm_FWVBErg_size(gdf[attr_colors_FWVB_Erg].astype(float)) * fac_ms_FWVB   
                                        ,color = cmap_FWVBErg(norm_FWVBErg_color(gdf[attr_colors_FWVB_Erg].astype(float))) 
                                        )  
                            
                        colors=cmap_FWVBErg(np.arange(cmap_FWVBErg.N))
                        attr_colors_FWVB_Erg_patches = [ mpatches.Patch(color=colors[i], label=attr_colors_FWVB_Erg_patches_fmt.format(value) ) for (i,value) in zip([0
                                                                                                                  ,cmap_FWVBErg.N-1],[gdf[attr_colors_FWVB_Erg].min(),                                                                                                                    
                                                                                                                                   gdf[attr_colors_FWVB_Erg].max()])
                                                                                                                                   ]        
                        attr_colors_FWVB_Erg_patches[0].set_label("{:s} ({:7.6f})".format(attr_colors_FWVB_Erg_patches[0].get_label(),gdf_FWVB[gdf_FWVB[attr_colors_FWVB_Erg].astype(float)>0][attr_colors_FWVB_Erg].min()))                            
                                                    
                    
                        gdf=gdf_FWVB[gdf_FWVB[attr_colors_FWVB_Erg].astype(float)<0]
                        if not gdf.empty:
                            gdf.plot(ax = ax
                                        ,zorder = 2 
                                        ,marker = '.'                           
                                        ,markersize = norm_FWVBErg_size(gdf[attr_colors_FWVB_Erg].astype(float).apply(lambda x: math.fabs(x))) * fac_ms_FWVB   
                                        ,color = cmap_FWVBErgNeg(norm_FWVBErgNeg_color(gdf[attr_colors_FWVB_Erg].astype(float).apply(lambda x: math.fabs(x)))) 
                                        )  
                            
                            
                            colors=cmap_FWVBErgNeg(np.arange(cmap_FWVBErgNeg.N))
                            attr_colors_FWVB_Erg_patches=attr_colors_FWVB_Erg_patches + [ mpatches.Patch(color=colors[i], label=attr_colors_FWVB_ErgNeg_patches_fmt.format(value) ) for (i,value) in zip([0
                                                                                                                      ,cmap_FWVBErgNeg.N-1],[-gdf[attr_colors_FWVB_Erg].max(),                                                                                                                    
                                                                                                                                       -gdf[attr_colors_FWVB_Erg].min()])
                                                                                                                                       ]        
                            attr_colors_FWVB_Erg_patches[2].set_label("{:s} ({:7.6f})".format(attr_colors_FWVB_Erg_patches[2].get_label(),gdf_FWVB[gdf_FWVB[attr_colors_FWVB_Erg].astype(float)<0][attr_colors_FWVB_Erg].max()))                            
                            
                            
                            
                            
                            
                    else:
                        
                        try:
                            colors=cmap_FWVBErg(norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float)))
                            pass
                        except ValueError:
                            colors=cmap_FWVBErg(.666)

                        gdf_FWVB.plot(ax = ax
                                    ,zorder = 2
                                    ,marker = '.'                           
                                    ,markersize = markersizes          
                                    ,color = colors#cmap_FWVBErg(norm_FWVBErg_color(gdf_FWVB[attr_colors_FWVB_Erg].astype(float))) 
                                    )    

                        if attr_colors_FWVB_Erg_patchValues == None:
                            attr_colors_FWVB_Erg_patchValues=np.arange(norm_min_FWVB_Erg,norm_max_FWVB_Erg+1,norm_diff_FWVB_Erg/4)                                                                                 
                        attr_colors_FWVB_Erg_patches = [ mpatches.Patch(color=cmap_FWVBErg(norm_FWVBErg_color(value)), label=attr_colors_FWVB_Erg_patches_fmt.format(value) ) for value in attr_colors_FWVB_Erg_patchValues]
                                                                                                                                                                                                                                                  

        
            ### KNOTen ###

            attr_colors_KNOT_Sach_patches=[]
            attr_colors_KNOT_Erg_patches=[]   
            attr_colors_KNOT_Erg_patchesNeg=[]                      
            if not gdf_KNOT.empty and attr_colors_KNOT_Erg != None:

                ##minValue=gdf_KNOT[attr_colors_KNOT_Erg].astype(float).min()
                
                # Erstellen der Colormaps 
                cmap_KNOTErg = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_KNOTErg', colors_KNOT_Erg, N = 256)     
                minValue=gdf_KNOT[attr_colors_KNOT_Erg].astype(float).min()
                if colors_KNOT_ErgNeg != None and minValue <0:                    
                    cmap_KNOTErgNeg = matplotlib.colors.LinearSegmentedColormap.from_list('cmap_KNOTErgNeg', colors_KNOT_ErgNeg, N = 256)                      
                # Funktionsweise Farbskalen nach Normierung: Werte ]0,1[ erhalten die Randfarben 0,1
            
                # Normierungen fuer Groesse und Farbe                
                if norm_min_KNOT_Erg == None:       
                    if colors_KNOT_ErgNeg != None and minValue <0:
                        norm_min_KNOT_Erg=0.
                    else:
                        norm_min_KNOT_Erg=gdf_KNOT[attr_colors_KNOT_Erg].astype(float).min()    
                if norm_max_KNOT_Erg == None:
                    if colors_KNOT_ErgNeg != None and minValue <0:
                        norm_max_KNOT_Erg=max(gdf_KNOT[attr_colors_KNOT_Erg].astype(float).max(),-gdf_KNOT[attr_colors_KNOT_Erg].astype(float).min())                        
                        norm_KNOTErg_color=plt.Normalize(vmin=norm_min_KNOT_Erg, vmax=gdf_KNOT[attr_colors_KNOT_Erg].astype(float).max())  
                        norm_KNOTErgNeg_color=plt.Normalize(vmin=norm_min_KNOT_Erg, vmax=-gdf_KNOT[attr_colors_KNOT_Erg].astype(float).min())  
                    else:
                        norm_max_KNOT_Erg=gdf_KNOT[attr_colors_KNOT_Erg].astype(float).max()                    
                else:                    
                    # Farbnormierungen
                    norm_KNOTErg_color=plt.Normalize(vmin=norm_min_KNOT_Erg, vmax=norm_max_KNOT_Erg)  
                    norm_KNOTErgNeg_color=plt.Normalize(vmin=norm_min_KNOT_Erg, vmax=norm_max_KNOT_Erg)                      
                norm_KNOTErg_Size = plt.Normalize(vmin=norm_min_KNOT_Erg, vmax=norm_max_KNOT_Erg)  
                # Werte ]vmin,vmax[ erhalten Werte ]0,1[ entsprechend der Normierung [0,1] mit [vmin,vmax]
                logger.debug("{:s}norm_KNOTErg_Size: norm_min_KNOT_Erg={:15.6f} norm_max_KNOT_Erg={:15.6f} .".format(logStr,norm_min_KNOT_Erg,norm_max_KNOT_Erg)) 

                # Groesse
                if fac_ms_KNOT == None:
                    if fac_ms_FWVB != None:
                        fac_ms_KNOT=fac_ms_FWVB
                    else:
                        fac_ms_KNOT=8000.                

                if colors_KNOT_ErgNeg != None and minValue <0:                    

                    gdf=gdf_KNOT[gdf_KNOT[attr_colors_KNOT_Erg].astype(float)>=0]
                    if not gdf.empty:
                        gdf.plot(ax = ax
                                    ,zorder = attr_colors_KNOT_Erg_zOrder 
                                    ,marker = '.'                           
                                    ,markersize = norm_KNOTErg_Size(gdf[attr_colors_KNOT_Erg].astype(float)) * fac_ms_KNOT   
                                    ,color = cmap_KNOTErg(norm_KNOTErg_color(gdf[attr_colors_KNOT_Erg].astype(float))) 
                                    )  
                        
                        colors=cmap_KNOTErg(np.arange(cmap_KNOTErg.N))
                        attr_colors_KNOT_Erg_patches = [ mpatches.Patch(color=colors[i], label=attr_colors_KNOT_Erg_patches_fmt.format(value) ) for (i,value) in zip([0
                                                                                                                 ,cmap_KNOTErg.N-1],[gdf[attr_colors_KNOT_Erg].min(),                                                                                                                    
                                                                                                                                  gdf[attr_colors_KNOT_Erg].max()])
                                                                                                                                  ]        
                        attr_colors_KNOT_Erg_patches[0].set_label("{:s} ({:7.6f})".format(attr_colors_KNOT_Erg_patches[0].get_label(),gdf_KNOT[gdf_KNOT[attr_colors_KNOT_Erg].astype(float)>0][attr_colors_KNOT_Erg].min()))
                            
                    
                    gdf=gdf_KNOT[gdf_KNOT[attr_colors_KNOT_Erg].astype(float)<0]
                    if not gdf.empty:
                        gdf.plot(ax = ax
                                    ,zorder = 2 
                                    ,marker = '.'                           
                                    ,markersize = norm_KNOTErg_Size(gdf[attr_colors_KNOT_Erg].astype(float).apply(lambda x: math.fabs(x))) * fac_ms_KNOT   
                                    ,color = cmap_KNOTErgNeg(norm_KNOTErgNeg_color(gdf[attr_colors_KNOT_Erg].astype(float).apply(lambda x: math.fabs(x)))) 
                                    )  

                        colors=cmap_KNOTErgNeg(np.arange(cmap_KNOTErgNeg.N))
                        attr_colors_KNOT_Erg_patchesNeg = [ mpatches.Patch(color=colors[i], label=attr_colors_KNOT_ErgNeg_patches_fmt.format(value) ) for (i,value) in zip([0
                                                                                                                 ,cmap_KNOTErgNeg.N-1],[-gdf[attr_colors_KNOT_Erg].max(),                                                                                                                    
                                                                                                                                  -gdf[attr_colors_KNOT_Erg].min()])
                                                                                                                                  ]     
                                                                                                                                        
                        attr_colors_KNOT_Erg_patchesNeg[0].set_label("{:s} ({:7.6f})".format(attr_colors_KNOT_Erg_patchesNeg[0].get_label(),gdf_KNOT[gdf_KNOT[attr_colors_KNOT_Erg].astype(float)<0][attr_colors_KNOT_Erg].max()))                                                                                                                                    
                        
                    attr_colors_KNOT_Erg_patches=attr_colors_KNOT_Erg_patches+attr_colors_KNOT_Erg_patchesNeg
                        
                    
                        
                        
                else:
                    
                    msFactor=norm_KNOTErg_Size(gdf_KNOT[attr_colors_KNOT_Erg].astype(float))   
                                        
                    if size_min_KNOT_Erg!= None and size_min_KNOT_Erg>0 and size_min_KNOT_Erg<1:   
                        if size_max_KNOT_Erg==None or size_max_KNOT_Erg>size_min_KNOT_Erg:
                            for idx,(index,row) in enumerate(gdf_KNOT.iterrows()):
                                if msFactor[idx] < size_min_KNOT_Erg:
                                        msFactor[idx]=size_min_KNOT_Erg                                         
                    if size_max_KNOT_Erg!= None and size_max_KNOT_Erg>0 and size_max_KNOT_Erg<1:
                        if size_min_KNOT_Erg==None or size_min_KNOT_Erg<size_max_KNOT_Erg:
                            for idx,(index,row) in enumerate(gdf_KNOT.iterrows()):
                                if msFactor[idx] > size_max_KNOT_Erg:
                                        msFactor[idx]=size_max_KNOT_Erg      
                                        
                    logger.debug(f"{logStr}min. ms Factor: {msFactor[np.logical_not(np.isnan(msFactor))].min()} * fac_ms_KNOT max. ms Factor: {msFactor[np.logical_not(np.isnan(msFactor))].max()} * fac_ms_KNOT")
                    
                    gdf_KNOT.plot(ax = ax
                                ,zorder = attr_colors_KNOT_Erg_zOrder 
                                ,marker = marker_KNOT_Erg                        
                                ,markersize = msFactor * fac_ms_KNOT      
                                ,color = cmap_KNOTErg(norm_KNOTErg_Size(gdf_KNOT[attr_colors_KNOT_Erg].astype(float))) 
                                )   
                    
                    # Legendeneinräge                                                     
                    if attr_colors_KNOT_Erg_patchValues == None:
                         norm_diff_KNOT_Erg=norm_max_KNOT_Erg-norm_min_KNOT_Erg
                         attr_colors_KNOT_Erg_patchValues=np.arange(norm_min_KNOT_Erg,norm_max_KNOT_Erg+1,norm_diff_KNOT_Erg/4)                 
                    attr_colors_KNOT_Erg_patches = [mpatches.Patch(
                                                     color=cmap_KNOTErg(norm_KNOTErg_Size(value))
                                                    ,label=attr_colors_KNOT_Erg_patches_fmt.format(value)
                                                    ) 
                                                    for value in attr_colors_KNOT_Erg_patchValues
                                                    ]
                    #attr_colors_KNOT_Erg_patches[-1].set_label("{:s} (max.: {:4.2f})".format(
                    #      attr_colors_KNOT_Erg_patches[-1].get_label()
                    #     ,gdf_KNOT[attr_colors_KNOT_Erg].max()
                    #    )
                    #    )
                        
                    



        
        
            if MapOn:
                if LabelsOnTop:
                    cx.add_basemap(ax, crs=gdf_ROHR.crs.to_string(), source = cx.providers.CartoDB.PositronNoLabels, zoom = Map_resolution)
                    cx.add_basemap(ax, crs=gdf_ROHR.crs.to_string(), source = cx.providers.CartoDB.PositronOnlyLabels, zoom = Map_resolution)
                else:
                    cx.add_basemap(ax, crs=gdf_ROHR.crs.to_string(), source = cx.providers.CartoDB.Positron, zoom = Map_resolution)

            ax.axis('off')

            if axTitle != None:
                ax.set_title(axTitle)


            #ax.legend(handles=attr_colors_ROHR_Sach_patches
            #          ,loc='upper left'
            #          ,facecolor='white', framealpha=.01)


        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise e        
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     
            return attr_colors_ROHR_Sach_patches,attr_colors_ROHR_Erg_patches,attr_colors_FWVB_Sach_patches,attr_colors_FWVB_Erg_patches,attr_colors_KNOT_Erg_patches



"""

"""

__version__='90.12.4.23.dev1'

import sys
import os
import logging
import pandas as pd
import re
import numpy as np

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

logger = logging.getLogger('PT3S')  

try:
    from PT3S import Rm
except ImportError:
    logger.debug("{0:s}{1:s}".format('ImportError: ','from PT3S import Rm - trying import Rm instead ... maybe pip install -e . is active ...')) 
    import Rm

try:
    from PT3S import Lx
except ImportError:
    logger.debug("{0:s}{1:s}".format('ImportError: ','from PT3S import Lx - trying import Lx instead ... maybe pip install -e . is active ...')) 
    import Lx

def addResVecToDfAlarmEreignisse(    
    dfAlarmEreignisse
   ,TCsLDSRes1=pd.DataFrame()
   ,TCsLDSRes2=pd.DataFrame()   
    ):
    """
    dfAlarmEreignisse:
        Nr:                 lfd. Nr (gebildet gem. NrBy und NrAsc)
        tA:                 Anfangszeit
        tE:                 Endezeit
        tD:                 Dauer des Alarms
        ZHKNR:              ZHKNR (die zeitlich 1., wenn der Alarm sich über mehrere ZHKNRn erstreckt)
        tD_ZHKNR:           Lebenszeit der ZHKNR; x-Annotationen am Anfang/Ende, wenn ZHK beginnt bei Res12-Anfang / andauert bei Res12-Ende; '-1', wenn Lebenszeit nicht ermittelt werden konnte
        ZHKNRn:             sortierte Liste der ZHKNRn des Alarms; eine davon ist ZHKNR; typischerweise die 1. der Liste
        LDSResBaseType:     SEG oder Druck
        OrteIDs:            OrteIDs des Alarms
        Orte:               Kurzform von OrteIDs des Alarms
        Ort:                der 1. Ort von Orte        
        SEGName:            Segment zu dem der 1. Ort des Alarms gehört
        DIVPipelineName:
        Voralarm:           ermittelter Vorlalarm des Alarms; -1, wenn kein Voralarm in Res12 gefunden werden konnte
        Type:               Typ des Kontrollraumns; z.B. p-p für vollständige Flussbilanzen; '', wenn kein Typ gefunden werden konnte
        Name:               Name des Bilanzraumes 
        NrSD:               lfd. Nr Alarm BaseType
        NrName:             lfd. Nr Alarm Name
        NrSEGName:          lfd. Nr Alarm SEGName
        AlarmEvent:         AlarmEvent-Objekt       
        BZKat:              Betriebszustandskategorie des Alarms

    Returns:
    dfAlarmEreignisse with 2 Cols added:              
        resIDBase:          die 1. OrtID von OrteIDs
        dfResVec:           der resVec des Alarms
    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
        

    try:            

        dfAlarmEreignisse['resIDBase']=dfAlarmEreignisse['OrteIDs'].apply(lambda x: x[0])
              
        ### Ergebnisvektor fuer alle Orte bestimmen
        dfResVecs={}
        dfResVecsLst=[]

        for indexAlarm, rowAlarm in dfAlarmEreignisse.iterrows():
       
            resIDBase=rowAlarm['resIDBase']

            if resIDBase in dfResVecs.keys():
                # resIDBase schon behandelt
                dfResVecsLst.append(dfResVecs[resIDBase])
                continue

            # Spalten basierend auf resIDBase bestimmen 
            ErgIDs=[resIDBase+ext for ext in Rm.ResChannelTypesAll]
            IMDIErgIDs=['IMDI.'+ID for ID in ErgIDs] # jede Spalte koennte anstatt "normal" als IMDI. vorhanden sein
            ErgIDsAll=[*ErgIDs,*IMDIErgIDs] 
           
            # Ergebnisspalten  
            if rowAlarm['LDSResBaseType']=='SEG':
                dfFiltered=TCsLDSRes1.filter(items=ErgIDsAll,axis=1)
            else:
                dfFiltered=TCsLDSRes2.filter(items=ErgIDsAll,axis=1)
        
            # Ergebnisspalten umbenennen
            colDct={}
            for col in dfFiltered.columns:            
                m=re.search(Lx.pID,col)
                colDct[col]=m.group('E')
            dfFiltered.name=resIDBase
            dfResVec=dfFiltered.rename(columns=colDct)        

            # Ergebnisvektor merken
            dfResVecs[resIDBase]=dfResVec

            dfResVecsLst.append(dfResVec)

            logger.debug("{:s}resIDBase: {:50s} Anzahl gefundener Spalten in TCsLDSRes: {:d}".format(logStr, resIDBase, len(dfResVec.columns.to_list())))   

        dfAlarmEreignisse['dfResVec']=dfResVecsLst


                                                                                                                          
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise e   

    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return dfAlarmEreignisse

def fGenAlarmVisTimeSpan(
                     tA
                    ,tE                   
                    # alle nachfolgenden Werte sollten dieselbe Einheit haben und in dieser Einheit ganzzahlig sein
                    ,timeSpan=pd.Timedelta('25 Minutes')
                    ,timeRoundStr='1T'
                    ,timeBoundaryMin=pd.Timedelta('3 Minutes')
                    ,timeRef='A' # Alarme die laenger als timeSpan sind: Anfang oder Ende werden mit timeSpan dargestellt
                    ):
                    """
                    erzeugt eine Zeitspanne in welcher ein Alarm Zwecks Analyse dargestellt wird

                    tA, tE sind Anfang und Ende des Alarms
                    bzw. definieren allgemein einen Zeitausschnitt

                    diese werden ab- (tA) bzw. aufgerundet (tE) mit timeRoundStr

                    zwischen den gerundeten Zeiten und tA/tE soll mindestens timeBoundaryMin liegen
                    wenn nicht, wird timeBoundaryMin auf tA/tE angewendet und dann wird gerundet 
                    dies wird jedoch nur angewendet, wenn die gerundeten Zeiten und tA/tE voneinander verschieden sind

                    timeSpan ist die gewuenschte minimale Zeitspanne

                    Alarme die kuerzer sind werden mit timeSpan dargestellt

                    Alarme die laenger sind: Anfang oder Ende wird mit timeSpan dargestellt
                    
                    """

                    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
                    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

                    try:

                        # Zeiten ab- und aufrunden
                        #    wenn tA bzw. tE bereits entsprechend rund sind, aendert die nachfolgende Ab- bzw. Aufrundung nichts                         
                        timeStart=tA.floor(freq=timeRoundStr)                                            
                        timeEnd=tE.ceil(freq=timeRoundStr) 
                       
                        # wenn gerundet wurde: gerundete Zeiten auf Mindestabstand pruefen
                        if tA-timeStart < timeBoundaryMin and tA!=timeStart:
                            timeStart=tA-timeBoundaryMin
                            timeStart= timeStart.floor(freq=timeRoundStr) 

                        if timeEnd-tE < timeBoundaryMin and tE!=timeEnd:
                            timeEnd=tE+timeBoundaryMin
                            timeEnd= timeEnd.ceil(freq=timeRoundStr) 

                        # gerundete Zeitspanne mit Zeitspanne vergleichen
                        timeLeft=timeSpan-(timeEnd-timeStart)

                        logger.debug("{:s}tA: {!s:s} timeStart: {!s:s} tE: {!s:s} timeEnd: {!s:s} timeLeft: {!s:s}".format(logStr,tA,timeStart,tE,timeEnd,timeLeft)) 

                        if timeLeft > pd.Timedelta('0 Seconds'): # die ggf. aufgerundete Zeitspanne ist kuerzer als timeSpan; timeSpan wird dargestellt
                            #; die Zeitspanne wird in der Mitte von timeSpan platziert; das führt zur eneuten Berechnung und Rundung von timeStart(timeEnd
                            timeStart=timeStart-timeLeft/2
                            timeStart= timeStart.floor(freq=timeRoundStr) 
                            timeEnd=timeEnd+timeLeft/2
                            timeEnd= timeEnd.ceil(freq=timeRoundStr) 
                        elif timeLeft == pd.Timedelta('0 Seconds'): # die ggf. aufgerundete Zeitspanne ist identisch mit timeSpan
                            pass
                        else:                    
                            # die ggf. aufgerundete Zeitspanne ist laenger als timeSpan; A oder E wird mit timeSpan wird dargestellt
                            if timeRef=='A':
                                timeM=tA.floor(freq=timeRoundStr) 
                            else:
                                timeM=tE.ceil(freq=timeRoundStr) 

                            timeStart=timeM-timeSpan/2
                            timeEnd=timeM+timeSpan/2
                          

                        if timeEnd-timeStart > timeSpan:
                            timeEnd=timeStart+timeSpan

                        ZeitbereichSel=timeEnd-timeStart

                        if ZeitbereichSel <= pd.Timedelta('1 Minutes'):
                                 bysecond=list(np.arange(0,60,1))
                                 byminute=None

                        elif ZeitbereichSel <= pd.Timedelta('3 Minutes'):
                                 bysecond=list(np.arange(0,60,5))
                                 byminute=None
                        elif ZeitbereichSel > pd.Timedelta('3 Minutes') and ZeitbereichSel <= pd.Timedelta('5 Minutes'):
                                 bysecond=list(np.arange(0,60,15))           
                                 byminute=None
                        elif ZeitbereichSel > pd.Timedelta('5 Minutes') and ZeitbereichSel <= pd.Timedelta('20 Minutes'):
                                 bysecond=list(np.arange(0,60,30))
                                 byminute=None
                        elif ZeitbereichSel > pd.Timedelta('20 Minutes') and ZeitbereichSel <= pd.Timedelta('30 Minutes'):
                                 bysecond=None
                                 byminute=list(np.arange(0,60,1))
                        elif ZeitbereichSel > pd.Timedelta('30 Minutes') and ZeitbereichSel <= pd.Timedelta('120 Minutes'):
                                 bysecond=None
                                 byminute=list(np.arange(0,60,3))
                        elif ZeitbereichSel > pd.Timedelta('120 Minutes') and ZeitbereichSel <= pd.Timedelta('180 Minutes'):
                                 bysecond=None
                                 byminute=list(np.arange(0,60,5))
                        elif ZeitbereichSel > pd.Timedelta('180 Minutes') and ZeitbereichSel <= pd.Timedelta('360 Minutes'):
                                 bysecond=None
                                 byminute=list(np.arange(0,60,10))
                        else:
                                 bysecond=None
                                 byminute=list(np.arange(0,60,20))


                    except Exception as e:
                        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
                        logger.error(logStrFinal) 
                        raise e   

                    finally:       
                        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
                        return timeStart, timeEnd, byminute, bysecond


                    #return timeStart, timeEnd, byminute, bysecond

def rptAlarms(     
     pdfErgFile='rptAlarms.pdf'

    ,figsize=Rm.DINA2q
    ,dpi=Rm.dpiSize

    ,dfAlarmStatistik=pd.DataFrame()   # 1 Zeile pro SEG; Spalten mit Alarm-Informationen zum SEG
    ,dfAlarmEreignisse=pd.DataFrame() # 1 Zeile pro Alarm; Spalten mit Informationen zum Alarm

    ,TCsLDSRes1=pd.DataFrame()   
    ,TCsLDSRes2=pd.DataFrame()       

    ,TCsLDSIn=pd.DataFrame() 
    ,TCsOPC=pd.DataFrame() 

    ,TCsSIDEvents=pd.DataFrame() 
    ,IDSetsDctAlNr={} # AlNr-Keyed dct with ColIDs

    ,timeSpanMin={} # AlNr-Keyed dct with Timespan-Para

    ,QDct={} # AlNr-Keyed dct with QDct-Para
    ,pDct={} # AlNr-Keyed dct with pDct-Para
    ,QDctOPC={} # AlNr-Keyed dct with pDct-Para
    ,pDctOPC={} # AlNr-Keyed dct with pDctOPC-Para

    ,attrsDct=Rm.attrsDct  

    ,plotOnlyAlNrn=None # zu Testzwecken; Liste der zu reportenden Alarme 

    ,*args
    ,**kwds
    ):
    """
          
    # ueber alle Segmente mit Alarmen

    # alle Alarme eines SEGS in der Reihenfolge ihrer Nummerierung (also i.d.R. zeitlich) hintereinander
    
    # jeden Alarm mit HYD und LDS in einem Bild darstellen  

    # es werden Einzelbilder und ein PDF erzeugt

    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

    try:   

        rptAlarmsResults={}
        # (gsHYD,gsLDS,pltLDSpQAndEventsResults,pltLDSErgVecResults)
                
        # PDF - DokAnfang
        pdf=PdfPages(pdfErgFile)
        (fileNameBase,ext)= os.path.splitext(pdfErgFile)

        # ueber alle Segmente 
      

        for indexSEG,rowSEG in dfAlarmStatistik.iterrows():

            strSEG="LfdNr {:2d} - {!s:3s}: {!s:15s}".format(
                    indexSEG+1
                   ,rowSEG.DIVPipelineName                           
                   ,rowSEG['SEGName']                                   
                 #  ,rowSEG['SEGResIDBase']
                  )

            if rowSEG['FörderZeitenAlAnz']==0 and rowSEG['RuheZeitenAlAnz']==0:
                logger.info("{:s}: FörderZeitenAlAnz=0 und RuheZeitenAlAnz=0".format(strSEG))     
                continue             
                
            # Segmente mit Alarmen ...

            # Alarmnummern
            AlNrn=sorted(rowSEG['FörderZeitenAlNrn']+rowSEG['RuheZeitenAlNrn'])
            #logger.info("{:s}: AlNrn: {!s:s}".format(strSEG, AlNrn))   


            # über alle Alarme des SEGs
            for idxAlarm,AlNr in enumerate(AlNrn):

                # der Alarm
                s=dfAlarmEreignisse[dfAlarmEreignisse['Nr']==AlNr].iloc[0]     

                titleStr="{:s}: AlNrn: {!s:s}: AlNr.: {:d} ({:s}: {:s})".format(strSEG, AlNrn, AlNr,s.LDSResBaseType,s.resIDBase)   

                if plotOnlyAlNrn != None:
                    if AlNr not in plotOnlyAlNrn:
                        logger.info("{:s}: nicht in plotOnlyAlNrn ...".format(titleStr))  
                        continue
                
                logger.info(titleStr)   

                # sein Ergebnisvektor
                resIDBase=s.resIDBase
                dfResVec=s.dfResVec

                # FIG
                fig=plt.figure(figsize=figsize,dpi=dpi)     


                # SEG- oder Druck-Alarm
                if s.LDSResBaseType=='SEG':                    
                    dfSegReprVec=dfResVec
                    dfDruckReprVec=pd.DataFrame()                    
                else:                    
                    dfSegReprVec=pd.DataFrame()
                    dfDruckReprVec=dfResVec

               
                timeStart, timeEnd, byminute, bysecond = fGenAlarmVisTimeSpan(s.tA,s.tE
                                                                           ,timeSpan=pd.Timedelta('25 Minutes')
                                                                           ,timeRoundStr='1T'
                                                                           ,timeBoundaryMin=pd.Timedelta('1 Minutes')
                                                                          )    
                xlims=[(timeStart, timeEnd)]
                byminute=byminute
                bysecond=bysecond
                vAreasX=[[(s.tA,s.tE)]]


                  
                gsHYD,gsLDS,pltLDSpQAndEventsResults,pltLDSErgVecResults=Rm.plotTimespans(     
                             xlims=xlims

                            ,sectionTitles=["{:s}".format(s.DIVPipelineName)]
                            ,sectionTitlesLDS=["tA={!s:s}".format(s.tA)]
                                                   
                            ,byminute=byminute
                            ,bysecond=bysecond
                            ,orientation='portrait'

                            ,vAreasX=vAreasX        
                            ,vLinesXLDS=[]
                            ,vAreasXLDS=[]   
                         
                            # --- Args Fct. ---:
                            ,TCsLDSIn=TCsLDSIn
                
                            ,TCsOPC=TCsOPC
                            ,TCsOPCScenTimeShift=pd.Timedelta('0 seconds') 

                            ,TCsSIDEvents=TCsSIDEvents.filter(items=IDSetsDctAlNr[s.Nr]['I'] if s.Nr in IDSetsDctAlNr.keys() else [])  
                            ,TCsSIDEventsTimeShift=pd.Timedelta('0 seconds') 
        
                            ,QDct=QDct[s.Nr]
                            ,pDct=pDct[s.Nr]
                            ,QDctOPC=QDctOPC[s.Nr] if s.Nr in QDctOPC.keys() else {}
                            ,pDctOPC=pDctOPC[s.Nr] if s.Nr in pDctOPC.keys() else {} 
                            ,attrsDct=attrsDct  
                            
                            #,fctsDct=fctsDct

                           
                         #   ,ylimp=(0,70)  
                         #   ,ylimQ=(-50,300)   
                         #   ,yGridSteps=14  
                        

                            # --- Args Fct. ---:            
                           ,dfSegReprVec=dfSegReprVec
                           ,dfDruckReprVec=dfDruckReprVec
        
                       
                   
                
                )

                rptAlarmsResults[AlNr]=(gsHYD,gsLDS,pltLDSpQAndEventsResults,pltLDSErgVecResults)

                fig.suptitle("{:s}: {!s:s} {!s:s}: ZHKName: {:s}: ZHKNR: {!s:s}".format(titleStr,s.Type,s.Voralarm,Rm.fCVDName(s.Name),s.ZHKNR),y=0.99)

                fig.tight_layout()

                # png                   
                fileName="{:s} {:2d} - {:s} {:s} {:s}.png".format(fileNameBase
                                                             ,int(rowSEG.Nr)+1
                                                             ,str(rowSEG.DIVPipelineName)
                                                             ,rowSEG['SEGName']
                                                             ,s.Orte[0]#resIDBase
                                                             )           
           
                # 
            
                fileNameAlarm="{:s} Nr {:d}.png".format(fileName.replace('.png',''),AlNr)                                  
                plt.savefig(fileNameAlarm)  
               
                plt.show()

                # PDF - Seite
                pdf.savefig(fig)            
                plt.close()  
    
        # PDF - DokEnde
        pdf.close()       
                                                                                                                   
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise e                       
    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return rptAlarmsResults

def plotDfAlarmEreignisseMitKat(    
     dfAlarmEreignisse=pd.DataFrame()    
    ,sortBy=[]
    ,replaceTup=('2021-','')
    ,replaceTuptD=('0 days','')
    ):
    """
    Returns the plt.table and the df behind it
    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
    


    try:                     
     
        df=dfAlarmEreignisse[['Nr'
                             ,'LDSResBaseType','Voralarm'
                             ,'Type'
                             ,'NrSD','tA','tE','tD'
                             ,'Orte'
                             ,'NrSEGName','SEGName','BZKat'
                             ,'Kategorie','Unterkat.','Information','Einfl. n. Rel.']].copy()

        df['tA']=df['tA'].apply(lambda x: str(x).replace(replaceTup[0],replaceTup[1]))
        df['tE']=df['tE'].apply(lambda x: str(x).replace(replaceTup[0],replaceTup[1]))    
        df['tD']=df['tD'].apply(lambda x: str(x).replace(replaceTuptD[0],replaceTuptD[1]))        

        df['Orte']=df['Orte'].apply(lambda x: str(x[0]))
        
        df['LDSResBaseType']=df.apply(lambda row: "{:s} {:s} - {:d}".format(row['LDSResBaseType'],row['Type'],row['Voralarm']),axis=1)
        df=df[['Nr','LDSResBaseType','NrSD','tA','tE','tD','NrSEGName','SEGName','Orte','BZKat','Kategorie','Unterkat.','Information','Einfl. n. Rel.']]
        df.rename(columns={'LDSResBaseType':'ResTyp - Voralarm'},inplace=True)
       
        df['NrSEGName (SEGName)']=df.apply(lambda row: "{!s:2s} ({!s:s})".format(row['NrSEGName'],row['SEGName']),axis=1)

        df=df[['Nr','ResTyp - Voralarm','NrSD','tA','tD'
               ,'Orte' 
               ,'BZKat'
               ,'NrSEGName (SEGName)','Kategorie','Unterkat.','Information','Einfl. n. Rel.']]

        df.rename(columns={'Orte':'ID'},inplace=True)       


        if sortBy!=[]:
            df=df.sort_values(by=sortBy)
       

        t=plt.table(cellText=df.values, colLabels=df.columns
                  
                    ,colWidths=[.03,.1 # Nr ResTyp-Voralarm
                                ,.04 # NrSD
                                ,.08,.08 # tA tD
                              
                                ,.1125,.07 # ID BZKat
                                
                                ,.14  # NrSEGName (SEGName)
                                ,.05 # Kat.
                                #,.055
                                #,.14,.1025] # 
                                #,.085
                                #,.155,.0575] # 
                                ,.105
                                ,.135,.0575] # 

                    , cellLoc='left'
                    , loc='center')    

        t.auto_set_font_size(False)
        t.set_fontsize(10)
                
        cols=df.columns.to_list()       
        colIdxNrSD=cols.index('NrSD')
        colIdxNrSEG=cols.index('NrSEGName (SEGName)')
        colIdxResTypVA=cols.index('ResTyp - Voralarm')
        colIdxUnterkat=cols.index('Unterkat.')

        cells = t.properties()["celld"]
        for cellTup,cellObj in cells.items():
             cellObj.set_text_props(ha='left')


             row,col=cellTup # row: 0 fuer Ueberschrift bei Ueberschrift; col mit 0
    
            
             if col == colIdxNrSD:
                 if row > 0:
                    if dfAlarmEreignisse.loc[row-1,'LDSResBaseType']=='SEG':
                        cellObj.set_text_props(backgroundcolor='lightsteelblue') 
                    else:
                        cellObj.set_text_props(backgroundcolor='plum') 

             elif col == colIdxNrSEG:
        
                if row==0:
                    continue

                if 'color' in dfAlarmEreignisse.columns.to_list():
                    color=dfAlarmEreignisse['color'].iloc[row-1]
                    cellObj.set_text_props(backgroundcolor=color)                       

             elif col == colIdxUnterkat:
        
                if row==0:
                    continue

                if dfAlarmEreignisse.loc[row-1,'Unterkat.']=='AT':
                   
                    cellObj.set_text_props(backgroundcolor='limegreen')   

                                  


             elif col == colIdxResTypVA and row > 0:
                pass


                        
                if dfAlarmEreignisse.loc[row-1,'Voralarm'] in [10]:
                        cellObj.set_text_props(backgroundcolor='sandybrown')  

                elif dfAlarmEreignisse.loc[row-1,'Voralarm'] in [4]:
                        cellObj.set_text_props(backgroundcolor='pink')   

                elif dfAlarmEreignisse.loc[row-1,'Voralarm'] in [3]:
                        cellObj.set_text_props(backgroundcolor='lightcoral')   


             else:
                 pass
                
           
       
            
    

        plt.axis('off')       
                                                                                                                   
    except Rm.RmError:
        raise            
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise Rm.RmError(logStrFinal)                       
    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return t,df



# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 14:21:15 2024

@author: wolters
"""

import logging
logger = logging.getLogger('PT3S') 

import sys

import re

import pandas as pd
import numpy as np

class SdfCsv():
    """Wrapper for a model defined by an SDF-CSV-File.
    """
    def __init__(self
                 ,csvFile='aSdfCsvFile.csv'
                 ,encoding='cp1252'
    ):
        """
        :param csvFile: a SDF-CSV-File
        :type csvFile: str
        :param encoding: encoding
        :type encoding: str    
                      
        .. note:: SIR 3S imports SDF-CSV-Files. However sometimes it is convenient to have a look at the original SDF-CSV-File content with Python. I.e. to explore the original hierarchical layer structure in comparison to the flat layer structure in the SIR 3S model. Or i.e. to check against the SIR 3S SDF-CSV-File import- and export-function result. In general, one can say that the SDF-CSV format for pipe network data (for pipe network models) is so widespread in Germany that it makes sense to have an SIR 3S affine function which maps the SDF-CSV-File to pandas dfs.
                   
            The returned object has the following structure:
        
                - dataFrames: dct with dfs
                - dataFrames['FEL']: a df with the definition of the columns of all types; i.e. 'KNO' is a type                
                - dataFrames['KNO']: a df with data for objects of type 'KNO'
                - dataFramesDesc: dct with descs; i.e. dataFramesDesc['KNO']: 'Knotendaten'
                
            The returned object has the following functions:

                - getHierarchicalLayernames: returns: {originalLayername:hierarchicalLayername, ...}                          
        """
        
        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
        
        try: 
            self.csvFile=csvFile
            self.encoding=encoding
            self.dataFrames,self.dataFramesDesc=self.__readSdfCsvFile()                  
                                           
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise e
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         

    def __readSdfCsvFile(self):
        """
        :return: dfs,dataFramesDesc              
        """
        
        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug(f"{logStr}Start.") 
        
        try:
            dataFrames={}
            dataFramesDesc={}
        
            with open(self.csvFile, encoding=self.encoding) as f:
                csvLines = f.readlines()
            
            csvLineNumbersWithFEL=[idx for idx,csvLine in enumerate(csvLines) if re.search('^FEL',csvLine) != None]
            #csvLinesWithFEL=[csvLine for csvLine in csvLines if re.search('^FEL',csvLine) != None]
            
            dfFEL=pd.read_csv(self.csvFile
                ,encoding=self.encoding
                ,sep=';'
                ,header=None
                ,index_col=False
                ,names=['CTY','OTY','XTY','ANAME','ATYPE','ALENGTH','APRECISION','ADESC','AUNIT']
                ,dtype={'CTY':str,'OTY':str,'XTY':str,'ANAME':str,'ATYPE':str,'ALENGTH':int,'APRECISION':int,'ADESC':str,'ANUNIT':str}           
                ,skiprows = lambda x: x not in  csvLineNumbersWithFEL
                ,on_bad_lines='warn'
               )
            cols=dfFEL.columns.to_list()
            dfFEL['lineNumber']=csvLineNumbersWithFEL
            dfFEL=dfFEL.filter(items=['lineNumber']+cols,axis=1)
            
            #nrows=10
            
            def fGetDype(row):
            
                match row['ATYPE']:                                
                    case _:
                        return str
                    
            def fGetConverter(row):
            
                match row['ATYPE']:
                    
                    case 'C':
                        return None
                    
                    case 'N':
                        
                        match row['APRECISION']:
                            
                            case 0:
                                return str.strip
                            case _:
                                return str.strip                                    
                    case _:
                        return None
            
            for OTY in sorted(dfFEL['OTY'].unique()):
                
                if OTY !=  'PRO':
                    pass
                    #continue
                    
                logger.info(f"###{OTY}###")
                
                dfFELOTY=dfFEL[dfFEL['OTY']==OTY]
                
                if dfFELOTY.empty:
                    logger.debug(f"No definition for type: {OTY}?!")
                    continue
                
                #logger.debug(f"{dfFELOTY.describe()}")
                logger.debug(f"{dfFELOTY.to_string()}")
                logger.debug(f"{dfFELOTY['ANAME'].to_list()}")
                logger.debug(f"{dfFELOTY['ATYPE'].unique()}")
                
                #print(sorted(dfFELOTY['ANAME'].to_list()))
                
                dtype={row['ANAME']:fGetDype(row) for index,row in dfFELOTY.iterrows() }
                #converters={row['ANAME']:fGetConverter(row) for index,row in dfFELOTY.iterrows() }
                #converters={key:value for (key,value) in converters.items() if value != None}
                
                csvLineNumbersWithOTY=[idx for idx,csvLine in enumerate(csvLines) if re.search('^'+OTY,csvLine) != None]   
                
                if len(csvLineNumbersWithOTY)==0:
                    logger.debug(f"No data for type: {OTY}?!")
                    continue
                                            
                logger.debug(f"csvLineNumbersWithOTY[0]:{csvLineNumbersWithOTY[0]}")
                descLine=csvLines[csvLineNumbersWithOTY[0]-4].strip()
                if re.search('^REM',descLine) != None:                
                    dataFramesDesc[OTY]=descLine
                else:
                    dataFramesDesc[OTY]=f"probably wrong estimated Desc: {descLine}"
                                
                #ncsvLineNumbersWithOTY=[idx for idx in range(len(csvLines)) if idx not in csvLineNumbersWithOTY]
                dfOTY=pd.read_csv(self.csvFile
                            ,encoding=self.encoding
                            ,sep=';'
                            ,decimal=','
                            ,header=None
                            ,index_col=False
                            ,names= ['OTY']+dfFELOTY['ANAME'].to_list()  
                            ,dtype=dtype
                            #,converters=converters
                            ,skiprows = lambda x: x not in  csvLineNumbersWithOTY
                            ,on_bad_lines='warn'
                            #,engine='c'
                            #,nrows=nrows
                )
                cols=dfOTY.columns.to_list()
                dfOTY['lineNumber']=csvLineNumbersWithOTY#[:nrows]
                dfOTY=dfOTY.filter(items=['lineNumber']+cols,axis=1)
                #dfs[OTY]=dfOTY
                
                for index,row in dfFELOTY.iterrows():
                    
                    match row['ATYPE']:
            
                        case 'C':
                            pass
            
                        case 'N':
            
                            match row['APRECISION']:
            
                                case 0:
                                    try:                                                
                                        dfOTY[row['ANAME']]=dfOTY[row['ANAME']].str.strip().astype(np.int64,errors='ignore')                                                
                                    except Exception as e:
                                        logger.info(f"{OTY}:")
                                        logStrFinal="Exception: Line: {:d}: {!s:s}: {:s}".format(sys.exc_info()[-1].tb_lineno,type(e),str(e))
                                        logger.error(logStrFinal)    
                                        
                                case _:
                                    dfOTY[row['ANAME']]=dfOTY[row['ANAME']].str.replace(',','.').astype(float)
                                    
                                    
                        case _:
                            pass
                                    
                dataFrames[OTY]=dfOTY  
                
                
            dataFrames['FEL']=dfFEL
            dataFramesDesc['FEL']='type FEL: Feldbeschreibungen (type definitions)'
            return dataFrames,dataFramesDesc
                
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise e
            
        finally:
            logger.debug(f"{logStr}_Done.")     
            
    def getHierarchicalLayernames(self,sep='#'):
        """
        :return: {originalLayername:(hierarchicalLayername,[IDPTop,IDPLev1,...,ID]) ...}              
        """
        
        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug(f"{logStr}Start.") 
        
        try:
            dct={}
            df=self.dataFrames['PRO'].copy(deep=True)
            
            uniqueTITELIndicator=df.groupby(by='TITEL').count()['lineNumber'].max()
            if uniqueTITELIndicator > 1:
                logger.info("TITEL not unique!")
                            
            df['PARENTID_']=df['PARENTID'].apply(lambda x: re.sub('^0+','',x.replace('$','').strip())).values
            def f(x):
                
                try:
                    ret = int(x)
                except:
                    if x == '':
                        ret=0
                    else:
                        ret = -1
                
                return ret   
            df['PARENTID_']=df['PARENTID_'].apply(lambda x: f(x))
            
            def fParentID(parentID,parentIDs):
                #logger.debug(f"parentID: {parentID} parentIDs: {parentIDs} ...")
                if parentID in [0,-1]: # exit
                    return
                else: # recurse   
                    for index, row in df.iterrows():
                        if row['ID']==parentID:
                            #logger.debug(f"parentID: {parentID} gefunden.")
                            parentIDs.append(row['ID'])
                            fParentID(row['PARENTID_'],parentIDs)
                            break            


            for index, row in df.iterrows():
                parentIDs=[]
                parentIDsRev=[]
                #logger.debug(f"Suchen nach Hierarchie für: {row['TITEL']} ID: {row['ID']} PARENTID: {row['PARENTID_']}")
                
                fParentID(row['PARENTID_'],parentIDs=parentIDs)
                #logger.debug(f"x parentIDs: {parentIDs}")
                
                layerName=''
                
                if len(parentIDs) > 0:                                        
                    parentIDsRev=parentIDs[::-1]                                                                               
                    for parentID in parentIDsRev: 
                        layerNameParent=df[df['ID']==parentID]['TITEL'].iloc[0]
                        if layerName=='':
                            layerName=layerNameParent
                        else:
                            layerName=layerName+sep+layerNameParent
                    
                    layerName=layerName+sep+row['TITEL']
                    #logger.debug(f"Hierarchie für: {row['TITEL']}: {layerName}")
                else:
                    layerName=row['TITEL']
                #logger.debug(f"hierarchischer Name für: {row['TITEL']}: {layerName}")
                
                if row['TITEL'] not in dct.keys():
                    pass
                else:
                    logger.info(f"TITEL not unique: {row['TITEL']}: hierarchicalLayername is set to hierarchicalLayername of lineNumber: {row['lineNumber']}")
                
                parentIDsRev.append(row['ID'])
                dct[row['TITEL']]=(layerName,parentIDsRev)
                
            return dct
           
                
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise e
            
        finally:
            logger.debug(f"{logStr}_Done.")     



import warnings

import os
import sys
import logging
import pandas as pd

import re

__version__='90.14.3.0.dev1'

try:              
    
    # -------------------------------------- 
    logger = logging.getLogger('PT3S')
    logger.addHandler(logging.NullHandler())      
    # -------------------------------------- 
    
    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)

    logger.debug("{0:s}{1:s}".format(logStr,'Start.'))     

    #H5
    warnings.filterwarnings('ignore',category=pd.io.pytables.PerformanceWarning) #your performance may suffer as PyTables will pickle object types that it cannot map directly to c-types 
    #warnings.filterwarnings('ignore',category=tables.exceptions.NaturalNameWarning) #\lib\site-packages\tables\path.py:100: NaturalNameWarning: object name is not a valid Python identifier: '3S'; it does not match the pattern ``^[a-zA-Z_][a-zA-Z0-9_]*$``; you will not be able to use natural naming to access this object; using ``getattr()`` will still work, though)
                                                                                 
except Exception as e:
    logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
    logger.error(logStrFinal)
    logger.error("{0:s}{1:s}".format(logStr,"the call to logging.exception('') follows ...")) 
    logging.exception('')  

finally:
    logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))




"""DB utilities (pyodbc based).

ODBC mit Access funktioniert nur mit 32-bit 
    solange/wenn kein 64-bit Access Treiber vorhanden.
ODBC mit Access funktioniert unter Windows 10 nur mit 2003er .mdb s 
    solange der neuere Treiber {Microsoft Access Driver (*.mdb, *.accdb)}
    unter Windows 10 nicht instanziert werden kann. 
    Der neuere Treiber wird zwar angezeigt im ODBC-Datenquellen-Administrator (32-Bit),
    man kann den Treiber dort aber auch fuer DSNs nicht instanzieren:
    "Das Betriebssystem ist momentan nicht zum Ausfuehren dieser Anwendung konfiguriert."

-v --sqlFile .\VSICS\V_VSICS.sql --mdbFile C:\3S\Modelle\FBG.mdb
"""

# Code in English.
# Beschreibungen und Kommentare (zumeist) in Deutsch.
import os
import shutil
import sys
import logging

logger = logging.getLogger('PT3S.UTILS')     
import argparse
import unittest
import doctest

import traceback

import re

import csv

import pyodbc 
# ...\Anaconda3_32\Scripts\conda.exe install -c anaconda pyodbc=3.0.10
# Ergebnis: ...\Anaconda3_32\Lib\site-packages\pyodbc-3.0.10-py3.5.egg-info

import pandas as pd

class DbError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)



def getMdbConStr(mdbFile):
    mdbFile = os.path.normpath(mdbFile)
    mdbFile = os.path.abspath(mdbFile)         
    

    Driver=[x for x in pyodbc.drivers() if x.startswith('Microsoft Access Driver')]
    if Driver == []:
                logStrFinal="{:s}{:s}: No Microsoft Access Driver!".format(logStr,accFile)     
                raise DbError(logStrFinal)  

    # ein Treiber ist installiert
    conStr=(
                r'DRIVER={'+Driver[0]+'};'
                r'DBQ='+mdbFile+';'
                )







    return conStr #'DRIVER={Microsoft Access Driver (*.mdb)}' + ';' + 'DBQ=' + mdbFile + ';' 

class Db:
    """DB utilities (pyodbc based).
    .
    """

    def __init__(self,conStr):
        """Datenbankverbindung herstellen.
        
        Arguments:
        conStr: Connection String
        Bsp. fuer conStr: 
            r"DRIVER={Microsoft Access Driver (*.mdb)};DBQ=.\Db.mdb;" 
            "DSN=xyz"
                Wenn DSN xyz eingerichtet ist.
        .
        """

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            self.conStr=conStr
            logger.debug("{0:s}conStr: {1:s}.".format(logStr,self.conStr)) 
            self.open()     
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))                         
        except:
            logging.exception('')  
        finally:
            pass
            
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

    def close(self):
        """Datenbankverbindung schliessen.
        .
        """

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            self.con.close()   
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))                         
        except:
            logging.exception('')  
        finally:
            pass
            
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

    def open(self):
        """Datenbankverbindung oeffnen.
        .
        """

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            self.con = pyodbc.connect(self.conStr)            
            self.cur = self.con.cursor()        
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))                         
        except:
            logging.exception('')  
        finally:
            pass
            
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

    def executeSqlCmd(self,sqlCmd):
        """Fuehrt SQL-Befehl aus.

        Arguments:
        sqlCmd: SQL-Befehl
        Bsp. fuer sqlCmd: 
            r"CREATE TABLE TDbTest(Col1 VARCHAR(254), Col2 VARCHAR(254))" 
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        #logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            #logger.debug("{0:s}SQL:{1:s} Ausfuehrung ...".format(logStr,sqlCmd))  
            self.cur.execute(sqlCmd)
            self.con.commit()
 
        except pyodbc.Error as e:  
            logger.debug("{0:s}SQL:{1:s} Error bei Ausfuehrung ...".format(logStr,sqlCmd))  
            raise pyodbc.Error(str(e))      
        
        except Exception as e:
            logger.debug("{0:s}SQL:{1:s} Error bei Ausfuehrung ...".format(logStr,sqlCmd))  
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise DbError(logStrFinal)     
        
        finally:
            pass
            #logger.debug("{0:s}... erfolgreich!".format(logStr))
                       
        #logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))    

    def testIfExists(self,Name):
        """Prueft ob Tabelle/View Name existiert.
           
        Arguments:         
        Name: Name der Tabelle / des Views
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        #logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            nameExists=None
            tableNames = [row.table_name for row in self.cur.tables()] # VIEWS sind auch TABLES hier
            if Name in tableNames:
                nameExists=True
            else:
                nameExists=False
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))     
        except:
            logging.exception('')  
        finally:
            #logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     
            return nameExists 
            
    def dropIfExists(self,Name):
        """Loescht Tabelle/View Name wenn diese existiert.
           
        Arguments:         
        Name: Name der Tabelle / des Views
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        #logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:
            if self.testIfExists(Name):
                dropSQL = "DROP TABLE {0:s}".format(Name) # VIEWS sind auch TABLES hier
                self.executeSqlCmd(dropSQL)

        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))     
        except:
            logging.exception('')  
        finally:
            pass
            
        #logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

    def dropIfExistsFromRegExp(self,regExp):
        """Loescht alle Tabellen/Views deren Name regExp matched.
           
        Arguments:       
        regExp: Regulaerer Ausdruck
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        #logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:           
            for tableName in [row.table_name for row in self.cur.tables()]: # VIEWS sind auch TABLES hier
                if re.search(regExp,tableName) != None:
                    self.dropIfExists(tableName)
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))        
        except:
            logging.exception('')  
        finally:
            pass
            
        #logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))    

    def executeFromSqlFile(self,sqlFile,sep=';',commentPattern='--.*'):
        """Fuehrt SQL-Befehle in sqlFile aus.
        
        Mehrere SQL-Befehle muessen mit sep in sqlFile voneinander getrennt werden.    
        sep muss (abgesehen von nachfolgenden Blanks und Kommentaren) wenn das letzte Zeichen einer Zeile sein       
        Arguments:
        sqlFile: Datei mit einem oder mehreren SQL-Befehlen.
        sep: 1 Zeichen mit dem die SQL-Befehle in sqlFile voneinander getrennt sind.
        commentPattern: regExp mit der die Kommentare erkannt werden sollen.

        SQL-Befehle mit CREATE ...:
             existieren die zu erzeugenden Tabellen/Views, dann werden sie zuvor geloescht.
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:            
            # sqlFile pruefen
            sqlFile = os.path.normpath(sqlFile) # frei von Redundanzen etc.
            if not os.path.isfile(sqlFile):
                logger.error("{0:s}{1:s} ist keine Datei oder die Datei existiert nicht.".format(logStr,sqlFile))   
                sys.exit()                        
            if not os.path.isabs(sqlFile):
                 sqlFile = os.path.abspath(sqlFile) # mit absolutem Pfad
            logger.debug("{0:s}SQL-Datei:{1:s}.".format(logStr,sqlFile)) 

            # sqlFile lesen
            with open(sqlFile,'r') as f:
                sqlFileLines = f.readlines()
            logger.debug("{0:s}SQL-Datei:{1:s} gelesen.".format(logStr,sqlFile)) 

            # Bereinigungen 
            sqlFileLinesStripped=[]
            for idx,line in enumerate(sqlFileLines):
                lineStripped=line

                lineStripped = lineStripped.lstrip() # fuehrende Leerzeichen loeschen
                lineStripped = lineStripped.rstrip() # endende Leerzeichen loeschen

                #logger.debug("{0:s}lineStripped 01:{1:s}.".format(logStr,lineStripped)) 

                regExpObj = re.compile(commentPattern)
                m = regExpObj.search(lineStripped)
                if m == None:
                    pass
                else:
                    lineStripped = lineStripped.replace(m.group(0),'')   # Kommentare loeschen
                    lineStripped = lineStripped.rstrip() # ggf. neu entstandene endende Leerzeichen loeschen

                if lineStripped == '' or lineStripped == '\n':
                    continue # keine reinen Leerzeilen weiter behandeln

                # Zeilen die jetzt nicht mit sep enden (== mehrzeilge Befehle) um 1 Blank ergaenzen
                # andernfalls koennten beim Verketten Syntaxfehler entstehen
                # sowie
                # Zeilen die mit sep enden mit sepExtension enden lassen
                # andernfalls koennte beim Splitten der Verkettung nach nur sep z.B. auch hier ';' faelschlicher eine Befehlstrennung vorgenommen werden
                sepExtension = '@'+'#'+sep # 
                regExpObj = re.compile(sep+'$')
                m = regExpObj.search(lineStripped)
                if m == None:
                    lineStripped=lineStripped+' '
                else:
                    lineStripped = lineStripped.replace(m.group(0),sepExtension) 
                #logger.debug("{0:s}lineStripped 02:{1:s}.".format(logStr,lineStripped)) 
                  
                sqlFileLinesStripped.append(lineStripped)

            # Bereinigungen fertig 
            for idx,lineStripped in enumerate(sqlFileLinesStripped):
                pass
                #logger.debug("{0:s}lineStripped Nr. {1:d}:{2:s}.".format(logStr,idx,lineStripped)) 
            
            # sqlFile in 1 string um diesen dann in SQL-Befehle zu splitten
            sqlCommandsAsOneString = ''
            for line in sqlFileLinesStripped:
                sqlCommandsAsOneString+=line
            #logger.debug("{0:s}sqlCommandsAsOneString:{1:s}.".format(logStr,sqlCommandsAsOneString)) 
            sqlCommandLst = sqlCommandsAsOneString.split(sepExtension)
            if sqlCommandLst[-1] == '':
                # der letzte Befehl ist "leer" weil auch die letzte Befehlssequenz mit sep abgeschlossen wurde 
                del sqlCommandLst[-1]

            # SQL-Befehle fertig
            for idx,sqlCommand in enumerate(sqlCommandLst):
                pass
                #logger.debug("{0:s}sqlCommand Nr. {1:d}:{2:s}.".format(logStr,idx,sqlCommand))                 
         
            # DROPs ermitteln
            regExpObj = re.compile('^CREATE(\s+)(\w+)(\s+)(\w+)')     
            dropCommandLst = ['DROP' + ' ' + regExpObj.search(sqlCommand.upper()).group(2) + ' ' + regExpObj.search(sqlCommand).group(4) 
                                 for sqlCommand in sqlCommandLst if regExpObj.search(sqlCommand.upper()) != None]
            for idx,dropCommand in enumerate(dropCommandLst):
                pass
                #logger.debug("{0:s}dropCommand Nr. {1:d}:{2:s}.".format(logStr,idx,dropCommand))    
          
            # DROPs absetzen
            regExpObj = re.compile('^DROP (\w+) (\w+)')     
            for sqlCommand in dropCommandLst:
                Name = regExpObj.search(sqlCommand).group(2)
                self.dropIfExists(Name)

            # SQL-Befehle absetzen
            for idx,sqlCommand in enumerate(sqlCommandLst):
                #logger.debug("{0:s}SQL Nr. {1:d}:{2:s}.".format(logStr,idx+1,sqlCommand))  
                self.cur.execute(sqlCommand)
                self.con.commit()
        
        except SystemExit:
            logger.error('{0:s}SytemExit Exeption.'.format(logStr))
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))        
        except:
            logging.exception('')  
        finally:
            pass
            
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))             

    def selectSmart(self,vt_Name):
        """Gibt den Inhalt von vt_Name wie folgt zurueck:
        rows [][]
        ,colNames []
        ,colTypes []
        ,colSizes []
        ,dicts [{}]
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:            
            rows=None   
            colNames=None
            colTypes=None
            colSizes=None
            dicts=None        

            if self.testIfExists(vt_Name) != True:
                logger.error("{0:s}Tabelle/View {1:s} existiert nicht.".format(logStr,vt_Name))         
                sys.exit(0)
               
            # Spalteninformationen auslesen
            rowsDef=self.cur.columns(vt_Name)
            colNames = [row.column_name for row in rowsDef]#self.cur.columns(vt_Name)]
            colTypes = [row.type_name   for row in rowsDef]#self.cur.columns(vt_Name)]
            colSizes = [row.column_size for row in rowsDef]#self.cur.columns(vt_Name)]
            selectStmt = "SELECT * FROM "+vt_Name
            #logger.debug("Execute {0}...".format(selectStmt))
            self.cur.execute(selectStmt)
            #logger.debug("Executed {0}.".format(selectStmt))
            
            rows = self.cur.fetchall()
            #logger.debug("{0:s} Fetched {1} rows with {2} columns.".format(logStr,len(rows), len(colNames)))
            dicts=[]
            for row in rows:
                dict={}
                for col,colName in zip(row,colNames):
                    dict[colName]=col
                dicts.append(dict)

        except SystemExit:
            logger.error('{0:s}SytemExit Exeption.'.format(logStr))
        except pyodbc.Error as e:
            logging.exception('{0:s} {1:s} {2:s}'.format(logStr, vt_Name, str(e)))
            raise pyodbc.Error('{0:s} {1:s} {2:s}'.format(logStr, vt_Name, str(e)))
        except Exception as e:
            logging.exception('{0:s} {1:s} {2:s}'.format(logStr, vt_Name, str(e)))  
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))    
            return rows,colNames,colTypes,colSizes,dicts

    def exportToCsvFile(self,Name,csvFile,csvFileMode='w',csvFileDelimiter=';',csvFileHeader=True, withPandas=True):
        """Exportiert den Inhalt von Tabelle/View Name in csvFile.
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:            
            # csvFile pruefen
            csvFile = os.path.normpath(csvFile) # frei von Redundanzen etc.
            if not os.path.isfile(csvFile):
                logger.info("{0:s}{1:s} ist keine Datei oder die Datei existiert nicht.".format(logStr,csvFile))                    
            if not os.path.isabs(csvFile):
                 csvFile = os.path.abspath(csvFile) # mit absolutem Pfad

            # Quelle pruefen
            if self.testIfExists(Name) != True:
                logger.error("{0:s}Tabelle/View {1:s} existiert nicht.".format(logStr,Name))         
                sys.exit(0)

            head,tail=os.path.split(csvFile)
            logger.debug("{0:s}csv-Datei: {1:s}: Tabelle/View {2:s} lesen und schreiben ....".format(logStr,tail,Name)) 

            if not withPandas:
                # Daten lesen
                #logger.debug("{0:s}Tabelle/View {1:s} lesen ...".format(logStr,Name))      
                rows,colNames,colTypes,colSizes,dicts = self.selectSmart(Name)

                # csvFile schreiben
                #logger.debug("{0:s}Tabelle/View {1:s} schreiben ...".format(logStr,Name))      
                with open(csvFile,csvFileMode) as f:
                    csv.register_dialect('exportToCsvFile','excel',delimiter=csvFileDelimiter,lineterminator='\n')
                    writer = csv.DictWriter(f,fieldnames=colNames,dialect='exportToCsvFile')
                    if csvFileHeader:
                        writer.writeheader()
                    writer.writerows(dicts)
            else:
                #logger.debug("{0:s}Tabelle/View {1:s} schreiben ...".format(logStr,Name))      
                df=pd.read_sql("select * from {:s}".format(Name),self.con)                
                df.to_csv(csvFile,sep=csvFileDelimiter,header=csvFileHeader,mode=csvFileMode,index=False)


        except SystemExit:
            logger.error('{0:s}SytemExit Exeption.'.format(logStr))
        except pyodbc.Error as e:  
            raise pyodbc.Error(str(e))        
        except Exception as e:
            logging.exception('{0:s} {1:s} {2:s}'.format(logStr, Name, str(e)))  
        finally:
            pass
            
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))            
        
    def readFromCsvFileWithHeader(self,csvFile,csvFileDelimiter=';'):
        """Liest den Inhalt von csvFile. Interpretiert die 1. Zeile als Spaltennamen. Gibt den Inhalt wie folgt zurueck:
        dicts [{}]
        ,rows [][] - Spalten in csvFile-Reihenfolge
        ,colNames  - SpaltenNamen in csvFile-Reihenfolge
        ,idxDct {} - Key: SpaltenName Value: SpaltenIndex im csvFile beginnend mit 0
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:            
            # csvFile pruefen
            csvFile = os.path.normpath(csvFile) # frei von Redundanzen etc.
            if not os.path.isfile(csvFile):
                logger.error("{0:s}{1:s} ist keine Datei oder die Datei existiert nicht.".format(logStr,csvFile))     
                sys.exit(0)               
            if not os.path.isabs(csvFile):
                 csvFile = os.path.abspath(csvFile) # mit absolutem Pfad
            logger.debug("{0:s}csv-Datei:{1:s}.".format(logStr,csvFile)) 

            #dicts
            csv.register_dialect('readFromCsvFile','excel',delimiter=csvFileDelimiter,lineterminator='\n')
            with open(csvFile) as f:
                reader=csv.DictReader(f,dialect='readFromCsvFile')

                # Spaltennummern von Spaltennamen
                colNames=reader.fieldnames
                #logger.debug('{0:s}Spaltennamen:{1:s}.'.format(logStr,str(colNames)))    
                idxDct={}
                for idx,colName in enumerate(colNames):
                    idxDct[colName]=idx

                dicts=[]
                for idx,dict in enumerate(reader): 
                    #logger.debug('{0:s}Dict von Zeile {1:d}:{2:s}.'.format(logStr,idx,str(dict)))    
                    dicts.append(dict)
            
            #rows - Spalten in csvFile-Reihenfolge
            rows=[]
            for idxRow,dict in enumerate(dicts):
                row=[]
                for idx,colName in enumerate(colNames):
                    row.append(dict[colName])
                rows.append(row)

        except SystemExit:
            logger.error('{0:s}SytemExit Exeption.'.format(logStr))     
        except:
            logging.exception('')  
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))    
            return dicts,rows,colNames,idxDct

    def importCsvFileWithHeaderToTable(self,tableName,csvFile,csvFileDelimiter=';'):
        """Liest den Inhalt von csvFile. Interpretiert die 1. Zeile als Spaltennamen. 
        Speichert den Inhalt in tableName.
        Wenn tableName nicht existiert wird tableName mit den erf. Spalten erzeugt.
        Wenn tableName existiert und Spalten fehlen !oder existierende Spalten nicht passen 
        wird tableName entsprechend um Spalten erweitert !oder die Spalten werden geaendert.  
        Existiert tableName muessen die existierenden _und referenzierten Spalten vom Typ Varchar sein 
        !sonst gehen die Daten dieser Spalten verloren.
        .
        """

        logStr = "{0:s}: ".format(sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

        try:            
            # csvFile pruefen
            csvFile = os.path.normpath(csvFile) # frei von Redundanzen etc.
            if not os.path.isfile(csvFile):
                logger.error("{0:s}{1:s} ist keine Datei oder die Datei existiert nicht.".format(logStr,csvFile))     
                sys.exit(0)               
            if not os.path.isabs(csvFile):
                 csvFile = os.path.abspath(csvFile) # mit absolutem Pfad
            logger.debug("{0:s}csv-Datei:{1:s}.".format(logStr,csvFile)) 

            # Daten lesen 
            dicts,rows,colNames,idxDct=self.readFromCsvFileWithHeader(csvFile,csvFileDelimiter=csvFileDelimiter)

            # benoetigte Spaltenbreiten feststellen
            colLengthNeeded={}
            # ueber alle Spalten
            for idx,colName in enumerate(colNames):
                colLengthNeeded[colName]=0
                for idxRow,row in enumerate(rows): # ueber alle Zeilen
                    cell=row[idx]
                    if len(cell)>colLengthNeeded[colName]:
                        colLengthNeeded[colName]=len(cell)

            # 
            self.dropIfExists(tableName)

            if self.testIfExists(tableName) != True:
                # Tabelle erzeugen
                logger.debug("{0:s}Tabelle {1:s} existiert nicht.".format(logStr,tableName))     
                createTableSQL = ",\n".join(["{0:s} VARCHAR({1:d})".format(colName,colLengthNeeded[colName]) for colName in colNames])
                createTableSQL = "CREATE TABLE {0:s} ({1:s})".format(tableName,createTableSQL)
                self.executeSqlCmd(createTableSQL)
            else:
                # Tabelle pruefen und ggf. anpassen
                pass

            # Daten importieren per INSERT
            wo = ",".join(["{0:s}".format(colName) for colName in colNames])
            for idxRow,dict in enumerate(dicts):
                was = ",".join(["'{0:s}'".format(dict[colName]) for colName in colNames])   
                insertSQL="INSERT INTO {0:s} \n({1:s}) \nVALUES({2:s})".format(tableName,wo,was)   
                self.executeSqlCmd(insertSQL)
              
        except SystemExit:
            logger.error('{0:s}SytemExit Exeption.'.format(logStr))     
        except:
            logging.exception('')  
        finally:
            pass

        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))        
            
class  test_Db(unittest.TestCase):        
        """Test von Db.
        .
        """
        def setUp(self):
            """Test von Db vorbereiten.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                                     
                # mdb
                self.mdbFile = None
                for dirpath, dirnames, filenames in os.walk('.'):
                    for filename in filenames:
                        if filename.endswith("Db.mdb"): # eine mdb-Datei suchen und verwenden
                            self.mdbFile = os.path.join(dirpath,filename)
                            break
                    if self.mdbFile != None:
                        break
                if self.mdbFile == None:
                    logger.error("{0:s}Die gewuenschte .mdb-Datei konnte nicht gefunden werden.".format(logStr)) 
                    sys.exit()
                else:
                    self.mdbFile = os.path.normpath(self.mdbFile)
                    self.mdbFile = os.path.abspath(self.mdbFile)
                    logger.debug("{0:s}Mdb-Datei: {1:s}.".format(logStr,self.mdbFile))                                                                                  
                    self.conStr = getMdbConStr(self.mdbFile)                                       
                    logger.debug("{0:s}conStr: {1:s}.".format(logStr,self.conStr))            
               
                # sql
                self.sqlFile = None
                for dirpath, dirnames, filenames in os.walk('.'):
                    for filename in filenames:
                        if filename.endswith("Db.sql"): # eine sql-Datei suchen und verwenden
                            self.sqlFile = os.path.join(dirpath,filename)
                            break
                    if self.sqlFile != None:
                        break
                if self.sqlFile == None:
                    logger.error("{0:s}Die gewuenschte .sql-Datei konnte nicht gefunden werden.".format(logStr)) 
                    sys.exit()
                else:
                    self.sqlFile = os.path.normpath(self.sqlFile)
                    self.sqlFile = os.path.abspath(self.sqlFile)
                    logger.debug("{0:s}Sql-Datei: {1:s}.".format(logStr,self.sqlFile))   

                # csv (Export)
                (root,ext) = os.path.splitext(self.sqlFile)
                self.csvFile = root+'.csv'
                logger.debug("{0:s}Csv-Datei: {1:s}.".format(logStr,self.csvFile))   

                # csv (Import)
                self.csvFileImport=self.csvFile.replace('Db.csv','DbImport.csv')
                logger.debug("{0:s}Csv-Datei Import: {1:s}.".format(logStr,self.csvFileImport))   

                # Name der Testtabelle
                self.tableName='T_Db'
                                
            except SystemExit:
                logger.error('{0:s}SytemExit Exeption.'.format(logStr))
            except:
                logging.exception('') 
            finally:
                pass
           
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))           

        #@unittest.skip(".")
        def test_01_InitDbNichtda(self):
            """Test von InitDbNichtda.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
               
            self.assertRaisesRegex(pyodbc.Error,'nicht gefunden',Db,self.conStr.replace('Db.mdb','DbNichtda.mdb'))

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        #@unittest.skip(".")
        def test_02_InitDb2016(self):
            """Test von InitDb2016.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            self.assertRaisesRegex(pyodbc.Error,'nicht von Ihrer Anwendung erkannt',Db,self.conStr.replace('Db.mdb','Db2016.mdb'))

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 
 
        #@unittest.skip(".")
        def test_03_Init(self):
            """Test von Init.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                self.assertIsInstance(obj.con,pyodbc.Connection,msg='con nicht vom Typ pyodbc.Connection')    
                self.assertIsInstance(obj.cur,pyodbc.Cursor,msg='cur nicht vom Typ pyodbc.Cursor')                              
            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))  
                raise pyodbc.Error                                
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 
 
        #@unittest.skip(".")
        def test_04_DropIfExists(self):
            """Test von DropIfExists.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                obj.dropIfExists(self.tableName)      
            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))     
                raise pyodbc.Error                                       
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        #@unittest.skip(".")
        def test_05_DropIfExistsFromRegExp(self):
            """Test von DropIfExistsFromRegExp.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                obj.dropIfExistsFromRegExp('^'+self.tableName)      
            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))     
                raise pyodbc.Error                                       
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        #unittest.skip(".")
        def test_06_ExecuteSqlCommand(self):
            """Test von ExecuteSqlCommand.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                obj.dropIfExists(self.tableName)      
                obj.executeSqlCmd("CREATE TABLE "+self.tableName+"(Col1 VARCHAR(254), Col2 VARCHAR(254))")

                obj.executeSqlCmd("INSERT INTO "+self.tableName+"(Col1, Col2) VALUES ('11','12')")
                obj.executeSqlCmd("INSERT INTO "+self.tableName+"(Col1, Col2) VALUES ('21','Bug')")
                obj.executeSqlCmd("UPDATE      "+self.tableName+" SET Col2='22' WHERE Col2='Bug'")

                obj.executeSqlCmd("INSERT INTO "+self.tableName+"(Col1) VALUES ('Bug')")
                obj.executeSqlCmd("DELETE FROM "+self.tableName+" WHERE Col1='Bug'")
                
                obj.cur.execute("SELECT * FROM "+self.tableName)
                rows = obj.cur.fetchall()
                for idxRow,row in enumerate(rows):
                    for idxCol, cell in enumerate(row):
                         logger.debug("{0:s}{1:s}".format(logStr,cell))
                
                self.assertEqual(rows[0][0],'11')
                self.assertEqual(rows[0][1],'12')
                self.assertEqual(rows[1][0],'21')
                self.assertEqual(rows[1][1],'22')

            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))     
                raise pyodbc.Error                                                                  
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        def test_07_ExecuteFromSqlFile(self):
            """Test von ExecuteFromSqlFile.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

          
            try:
                obj=Db(self.conStr)
                # Testtabelle T_Db existiert noch 2x2 aus vorherigen Tests
                obj.executeFromSqlFile(self.sqlFile)

                obj.cur.execute("SELECT * FROM "+self.tableName)
                rows = obj.cur.fetchall()
                for idxRow,row in enumerate(rows):
                    for idxCol, cell in enumerate(row):
                         logger.debug("{0:s}{1:s}".format(logStr,cell))
                
                self.assertEqual(rows[0][0],'11')
                self.assertEqual(rows[0][1],'12')
                self.assertEqual(rows[1][0],'21')
                self.assertEqual(rows[1][1],'22FileDb.sql')

            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))    
                raise pyodbc.Error                                                                   
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        def test_08_ExportToCsvFile(self):
            """Test von ExportToCsvFile und ReadFromCsvFile.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                # Testtabelle T_Db existiert noch 2x2 aus vorherigen Tests
                obj.exportToCsvFile(self.tableName,self.csvFile,csvFileDelimiter=';')
                obj.exportToCsvFile(self.tableName,self.csvFile,csvFileMode='a',csvFileHeader=None)

                dicts,rows,colNames,idxDct=obj.readFromCsvFileWithHeader(self.csvFile,csvFileDelimiter=';')
               
                self.assertEqual(dicts[0],dicts[2])
                self.assertEqual(dicts[1],dicts[3])
                self.assertEqual('22FileDb.sql',dicts[1]['Col2'])

                self.assertEqual(rows[0],rows[2])
                self.assertEqual(rows[1],rows[3])
                self.assertEqual('22FileDb.sql',rows[1][1])
                
            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))    
                raise pyodbc.Error                                                                   
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 

        def test_09_ImportCsvFileWithHeaderToTable(self):
            """Test von ImportCsvFileWithHeaderToTable.
            .
            """

            logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

            try:
                obj=Db(self.conStr)
                obj.importCsvFileWithHeaderToTable(self.tableName,self.csvFileImport,csvFileDelimiter=';')
                rows,colNames,colTypes,colSizes,dicts=obj.selectSmart(self.tableName)
                self.assertEqual('22FileDbImport.csv',rows[1][1])
                
            except pyodbc.Error as e:
                logger.error('{0:s}Unerwartete Exception:{1:s} erhalten:{2:s}.'.format(logStr,'pyodbc.Error',str(e)))    
                raise pyodbc.Error                                                                   
            finally:
                pass

            logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 
              
if __name__ == "__main__":
    """.
    .
    """

    try:              
        # Logfile
        head,tail = os.path.split(__file__)
        file,ext = os.path.splitext(tail)
        logFileName = os.path.normpath(os.path.join(head,os.path.normpath('./Test'))) # Ablage des Logfiles auf Unterverzeichnis Test
        logFileName = os.path.join(logFileName,file + '.log') # Dateiname des Logfiles wie diese Datei - nur mit .log statt .py
        
        loglevel = logging.INFO
        logging.basicConfig(filename=logFileName
                            ,filemode='w'
                            ,level=loglevel
                            ,format="%(asctime)s ; %(name)-60s ; %(levelname)-7s ; %(message)s")    

        fileHandler = logging.FileHandler(logFileName)        
        logger.addHandler(fileHandler)

        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logging.Formatter("%(levelname)-7s ; %(message)s"))
        consoleHandler.setLevel(logging.INFO)
        logger.addHandler(consoleHandler)

        logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
                                      
        # Argumente parsen      
        parser = argparse.ArgumentParser()
        parser.add_argument("--sqlFile",type=str, help="file with SQL-Commands (.sql-File) to be executed in Access Database")    
        parser.add_argument("--mdbFile",type=str, help="Access Database (.mdb-File)")                
        parser.add_argument("-v","--verbose", help="Debug Messages On", action="store_true")               
        args = parser.parse_args()

        if args.verbose:           
            logger.setLevel(logging.DEBUG)                   
        logger.debug("{0:s}{1:s}{2:s}".format(logStr,'Start. Argumente:',str(sys.argv))) 

        if args.sqlFile and args.mdbFile:
            # Scripting-Einsatz
            sqlFile = args.sqlFile 
            mdbFile = args.mdbFile
            conStr=getMdbConStr(mdbFile)
            obj = Db(conStr)
            obj.executeFromSqlFile(sqlFile)
            obj.close()
            
        else:
            # Selbsttest
            # WD: .\UTILS
            # Arguments: -v
            unittest.main(verbosity=2)
                                                     
    except:
        logger.error("{0:s}{1:s}".format(logStr,'logging.exception!')) 
        logging.exception('')  
    finally:
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 
        sys.exit(0)






"""
# Module Test:
# -q -m 1 --testDir testdata
# all Single Tests:
# -q -m 0 -s Dm  
# ---
# SETUP
# ---
import os
import time
import logging
logger = logging.getLogger('PT3S.Dm')  
# ---
# path
# ---
if __name__ == "__main__":
  try:      
     dummy=__file__
     logger.debug("{0:s}{1:s}{2:s}".format('DOCTEST: __main__ Context: ','path = os.path.dirname(__file__)'," .")) 
     path = os.path.dirname(__file__)
  except NameError:    
     logger.debug("{0:s}{1:s}{2:s}".format('DOCTEST: __main__ Context: ',"path = '.' because __file__ not defined and: "," from Dm import Dm")) 
     path = '.'
     from Dm import Dm
else:
  logger.debug("{0:s}{1:s}{2:s}{3:s}".format('DOCTEST: Not __main__ Context: ','__name__: ',__name__,"path = '.'")) 
  path = '.'
try:
  from PT3S import Am
except ImportError:
  logger.debug("{0:s}{1:s}".format("DOCTEST: ImportError: from PT3S import Am: ","- trying import Am instead ... maybe pip install -e . is active ..."))  
  import Am
# ---
# testDir
# ---
# globs={'testDir':'testdata'}
try:
   dummy= testDir
except NameError:
   testDir='testdata' 
# ---
# Init
# ---
accFile=os.path.join(path,os.path.join(testDir,'DHNetwork.mdb')) 
am=Am.Am(accFile=accFile)
accFile2=r'C:\\3s\\Projekte\\19.137 - Actemium - FBG LDS AP13\\04 - Versionen\\Version80_installiert_IO_20201216\\MDBDOC\\FBG.mdb'
am2=Am.Am(accFile=accFile2)
#accFile3=r'c:\\3s\\Projekte\\20.175 - SWM Geomare\\08 Bearbeitung Freimann\\Freimann.mdb'
#am3=Am.Am(accFile=accFile3)
"""

__version__='90.12.4.18.dev1'



import os
import sys


import re
import pandas as pd
import numpy as np
import warnings
import tables

import h5py
import time

import base64
import struct

import logging

import glob


import math

import argparse
import unittest
import doctest

# ---
# --- PT3S Imports
# ---
logger = logging.getLogger('PT3S')  
if __name__ == "__main__":
    logger.debug("{0:s}{1:s}".format('in MODULEFILE: __main__ Context:','.')) 
else:
    logger.debug("{0:s}{1:s}{2:s}{3:s}".format('in MODULEFILE: Not __main__ Context: ','__name__: ',__name__," .")) 

class DmError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def f_HelperBVBZ(
    con
   ,BV
   ,BZ   
    ):
    """
    Returns:
         df,dfBV,dfBZ
    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

    try:   

        df=pd.DataFrame()
            
        sql='select * from '+BV
        try:
            dfBV=pd.read_sql(sql,con)
        except pd.io.sql.DatabaseError as e:
            logStrFinal="{0:s}sql: {1:s}: Fehler?!".format(logStr,sql) 
            raise DmError(logStrFinal)        

        sql='select * from '+BZ
        try:
            dfBZ=pd.read_sql(sql,con)
        except pd.io.sql.DatabaseError as e:
            logStrFinal="{0:s}sql: {1:s}: Fehler?!".format(logStr,sql) 
            raise DmError(logStrFinal) 

        
        df=pd.merge(dfBZ
                        ,dfBV                                                          
                        ,left_on=['fk']
                        ,right_on=['pk']
                        ,suffixes=('_BZ',''))     
            
        newCols=df.columns.to_list()
        df=df.filter(items=[col for col in dfBV.columns.to_list()]+[col for col in newCols if col not in dfBV.columns.to_list()])
                                                                                                                  
    except DmError:
        raise            
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise DmError(logStrFinal)                       
    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return df,dfBV,dfBZ

def f_HelperDECONT(   
    df
   ,dfViewModelle 
   ,dfCONT
    ):
    """
    Returns:
        df
    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 

    try:   
               
        cols=df.columns.to_list()
        
        if 'fkDE_BZ' in cols:
            dfOrig=df
            df=pd.merge(df
                        ,dfViewModelle                                                        
                        ,left_on=['fkDE_BZ']
                        ,right_on=['fkBZ']
                        ,suffixes=('','_VMBZ'))   
            if df.empty:
                logger.debug("{0:s}{1:s}".format(logStr,'fkDE_BZ ist vmtl. kein BZ-Schluessel, da es sich vmtl. um keine BZ-Eigenschaft handelt sondern um eine BV-Eigenschaft; Spalten werden umbenannt und es wird nach BV-DE gesucht ...')) 
                renDct={col:col.replace('_BZ','_BV') for col in df.columns.to_list() if re.search('_BZ$',col)!=None}
                dfOrig.rename(columns=renDct,inplace=True)
                
                if 'fkDE' in cols:                    
                    logger.debug("{0:s}{1:s}".format(logStr,'fkDE ist auch in den Spalten ...')) 

                    df=pd.merge(dfOrig
                                ,dfViewModelle                                                        
                                ,left_on=['fkDE']
                                ,right_on=['fkBASIS']
                                ,suffixes=('','_VMBASIS')
                                ,how='left')         
                    df=pd.merge(df
                                ,dfViewModelle                                                        
                                ,left_on=['fkDE']
                                ,right_on=['fkVARIANTE']
                                ,suffixes=('','_VMVARIANTE')
                                ,how='left')    


                else:
                    logger.debug("{0:s}{1:s}".format(logStr,'fkDE ist nicht in den Spalten?!')) 
             
        elif 'fkDE' in cols: # (und 'fkDE_BZ' gibt es nicht)
            df=pd.merge(df
                        ,dfViewModelle                                                        
                        ,left_on=['fkDE']
                        ,right_on=['fkBASIS']
                        ,suffixes=('','_VMBASIS')
                        ,how='left')         
            df=pd.merge(df
                        ,dfViewModelle                                                        
                        ,left_on=['fkDE']
                        ,right_on=['fkVARIANTE']
                        ,suffixes=('','_VMVARIANTE')
                        ,how='left')       
            
        if 'fkCONT' in cols:
            df=pd.merge(df
                        ,dfCONT.add_suffix('_CONT')                                                        
                        ,left_on=['fkCONT']
                        ,right_on=['pk_CONT']
                        #,suffixes=('','_CONT')
                        )   
                                                                                                                          
    except DmError:
        raise            
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise DmError(logStrFinal)                       
    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return df

def f_HelperVKNO(   
    dfKNOT
   ,dfVKNO   
    ):
    """
    Returns:
        df
    """

    logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
    logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 


    logger.debug("{:s}dfKNOT: {!s:s}".format(logStr,dfKNOT.columns.to_list())) 
    logger.debug("{:s}dfVKNO: {!s:s}".format(logStr,dfVKNO.columns.to_list())) 

    try:                          
        df=pd.merge(dfKNOT,dfVKNO.add_suffix('_VKNO'),left_on='pk',right_on='fkKNOT_VKNO',how='left'
                  # ,suffixes=('','_VKNO')
                    )
                                                                                                                          
    except DmError:
        raise            
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise DmError(logStrFinal)                       
    finally:       
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))         
        return df
   

class Dm():
    """Processing Pandas Base Views.

    Args:
        * ...
           
    Attributes:
       
    Raises:
        DmError
    """


               
    def __init__(self):

        logStr = "{0:s}.{1:s}: ".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        logger.debug("{0:s}{1:s}".format(logStr,'Start.')) 
        

        try:
            pass
           









           
                          
        except Exception as e:
            logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
            logger.error(logStrFinal) 
            raise AmError(logStrFinal)              
        finally:
            logger.debug("{0:s}{1:s}".format(logStr,'_Done.'))     

  
if __name__ == "__main__":
    """
    Run Tests.
    """

    try:      
        
        logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
                                      
        # Arguments      
        parser = argparse.ArgumentParser(description='Run Tests.'
        ,epilog='''
        UsageExamples: 



        '''                                 
        )

        group = parser.add_mutually_exclusive_group()                                
        group.add_argument("-v","--verbose", help="Debug Messages On (default)", action="store_true",default=True)      
        group.add_argument("-q","--quiet", help="Debug Messages Off", action="store_true")           

        parser.add_argument('--testDir',type=str,default='testdata',help="value for global 'testDir' i.e. testdata (default)")
    
                                 
        parser.add_argument("-m","--moduleTest", help="execute the Module Doctest On/Off: -m 1 (default)", action="store",default='1')      
        parser.add_argument("-s","--singleTest", help='execute single Doctest: Exp.: -s  "^Dm.": all Doctests in Module Rm are executed - but not the Module Doctest (which is named Dm)'
                            ,action="append"
                            ,default=[])      
        parser.add_argument("-x","--singleTestNO", help='execute NOT single Doctest: Exp.: -s  "^Dm.": NO Doctests in Module Rm are executed - but not the Module Doctest (which is named Dm)'
                            ,action="append"
                            ,default=[])               


        parser.add_argument("-l","--logExternDefined", help="Logging (File etc.) ist extern defined (default: False)", action="store_true",default=False)      


        args = parser.parse_args()


        class LogStart(Exception):
            pass

        try:
            logStr = "{0:s}.{1:s}: ".format(__name__, sys._getframe().f_code.co_name)
            if args.logExternDefined:  
                logger = logging.getLogger('PT3S')  
                logStr=logStr+" (Logging extern defined) "
            else:                
                logFileName = 'PT3S.log' 
        
                loglevel = logging.INFO
                logging.basicConfig(filename=logFileName
                                    ,filemode='w'
                                    ,level=loglevel
                                    ,format="%(asctime)s ; %(name)-60s ; %(levelname)-7s ; %(message)s")    

                fileHandler = logging.FileHandler(logFileName)        
                logger.addHandler(fileHandler)

                consoleHandler = logging.StreamHandler()
                consoleHandler.setFormatter(logging.Formatter("%(levelname)-7s ; %(message)s"))
                consoleHandler.setLevel(logging.INFO)
                logger.addHandler(consoleHandler)
                          
            raise LogStart
        except LogStart:   
            if args.verbose:  # default         
                logger.setLevel(logging.DEBUG)  
            if args.quiet:    # Debug Messages are turned Off
                logger.setLevel(logging.ERROR)  
                args.verbose=False            
            logger.debug("{0:s}{1:s}".format(logStr,'Start.'))             
        else:
            pass
                                            
        logger.debug("{0:s}{1:s}{2:s}".format(logStr,'Argumente:',str(sys.argv))) 
        #logger.debug("{0:s}{1:s}{2:s}".format(logStr,'testDir: ',args.testDir)) 


        if args.moduleTest == '1':            
            logger.info("{0:s}{1:s}{2:s}".format(logStr,'Start unittests (by DocTestSuite...). testDir: ',args.testDir)) 
            dtFinder=doctest.DocTestFinder(recurse=False,verbose=args.verbose) # recurse = False findet nur den Modultest
            suite=doctest.DocTestSuite(test_finder=dtFinder #,setUp=setUpFct
                                   ,globs={'testDir':args.testDir                                           
                                           })   
            unittest.TextTestRunner().run(suite)
          
        if len(args.singleTest)>0:


 
            dtFinder=doctest.DocTestFinder(verbose=args.verbose)
            
            logger.debug("{:s}singleTests suchen in Mx ...".format(logStr)) 
            dTests=dtFinder.find(Dm,globs={'testDir':args.testDir}) 

            
            # gefundene Tests mit geforderten Tests abgleichen
            testsToBeExecuted=[]
            for expr in args.singleTest:
                logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Searching in Tests found for Expr     TBD',expr.strip("'")))                
                testsToBeExecuted=testsToBeExecuted+[test for test in dTests if re.search(expr.strip("'"),test.name) != None]     
            logger.debug("{0:s}singleTests: {1:s}: {2:s}".format(logStr,'    TBD',str(sorted([test.name for test in testsToBeExecuted]))))                   

            testsNotToBeExecuted=[]
            for expr in args.singleTestNO:
                logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Searching in Tests found for Expr NOT TBD',expr.strip("'")))      
                testsNotToBeExecuted=testsNotToBeExecuted+[test for test in testsToBeExecuted if re.search(expr.strip("'"),test.name) != None]       
            logger.debug("{0:s}singleTests: {1:s}: {2:s}".format(logStr,'NOT TBD',str(sorted([test.name for test in testsNotToBeExecuted]))))    

            # effektiv auszuführende Tests 
            testsToBeExecutedEff=sorted(set(testsToBeExecuted)-set(testsNotToBeExecuted),key=lambda test: test.name)
            
            dtRunner=doctest.DocTestRunner(verbose=args.verbose) 
            for test in testsToBeExecutedEff:                      
                    logger.debug("{0:s}singleTests: {1:s}: {2:s} ...".format(logStr,'Running Test',test.name)) 
                    dtRunner.run(test)                                        



                                
    except SystemExit:
        pass                                              
    except Exception as e:
        logStrFinal="{:s}Exception: Line: {:d}: {!s:s}: {:s}".format(logStr,sys.exc_info()[-1].tb_lineno,type(e),str(e))
        logger.error(logStrFinal) 
        raise AmError(logStrFinal)   
    finally:
        logger.debug("{0:s}{1:s}".format(logStr,'_Done.')) 


        sys.exit(0)

